"""Enumeration of data delivery types."""
#   Copyright 2021 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from enum import Enum


class DataDelivery(Enum):
    """
    Enumeration of available data delivery types.

    A device's data delivery mode is either an always connected device (PUSH),
    or a device that only periodically establishes connection and
    then subsequently checks if there are any pending messages
    that are intended for it (PULL).
    """

    PULL = "PULL"
    PUSH = "PUSH"



"""Deserialize messages received in WolkAbout Protocol format."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import json
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

from wolk import logger_factory
from wolk.interface.message_deserializer import MessageDeserializer
from wolk.model.device import Device
from wolk.model.file_transfer_package import FileTransferPackage
from wolk.model.message import Message


class WolkAboutProtocolMessageDeserializer(MessageDeserializer):
    """
    Deserialize messages received from the WolkAbout IoT Platform.

    :ivar logger: Logger instance issued by wolk.LoggerFactory
    :vartype logger: logging.Logger
    """

    PLATFORM_TO_DEVICE = "p2d/"
    CHANNEL_DELIMITER = "/"

    TIME = "time"
    PARAMETERS = "parameters"
    FEED_VALUES = "feed_values"

    FILE_BINARY = "file_binary_response"
    FILE_DELETE = "file_delete"
    FILE_PURGE = "file_purge"
    FILE_LIST = "file_list"
    FILE_UPLOAD_ABORT = "file_upload_abort"
    FILE_UPLOAD_INITIATE = "file_upload_initiate"
    FILE_URL_ABORT = "file_url_download_abort"
    FILE_URL_INITIATE = "file_url_download_initiate"

    FIRMWARE_ABORT = "firmware_update_abort"
    FIRMWARE_INSTALL = "firmware_update_install"

    def __init__(self, device: Device) -> None:
        """
        Create inbound topics from device key.

        :param device: Contains device key used for inbound topics
        :type device: Device
        """
        self.logger = logger_factory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )
        self.logger.debug(f"{device}")
        self.key = device.key

        self.time_topic = self._form_topic(self.TIME)
        self.feed_values_topic = self._form_topic(self.FEED_VALUES)
        self.parameters_topic = self._form_topic(self.PARAMETERS)

        self.file_binary_topic = self._form_topic(self.FILE_BINARY)
        self.file_delete_topic = self._form_topic(self.FILE_DELETE)
        self.file_purge_topic = self._form_topic(self.FILE_PURGE)
        self.file_list = self._form_topic(self.FILE_LIST)
        self.file_upload_abort_topic = self._form_topic(self.FILE_UPLOAD_ABORT)
        self.file_upload_initiate_topic = self._form_topic(
            self.FILE_UPLOAD_INITIATE
        )
        self.file_url_abort_topic = self._form_topic(self.FILE_URL_ABORT)
        self.file_url_initiate_topic = self._form_topic(self.FILE_URL_INITIATE)

        self.firmware_abort_topic = self._form_topic(self.FIRMWARE_ABORT)
        self.firmware_install_topic = self._form_topic(self.FIRMWARE_INSTALL)

        self.inbound_topics = [
            self.time_topic,
            self.feed_values_topic,
            self.parameters_topic,
            self.file_binary_topic,
            self.file_delete_topic,
            self.file_purge_topic,
            self.file_list,
            self.file_upload_abort_topic,
            self.file_upload_initiate_topic,
            self.file_url_abort_topic,
            self.file_url_initiate_topic,
            self.firmware_abort_topic,
            self.firmware_install_topic,
        ]
        self.logger.debug(f"inbound topics: {self.inbound_topics}")

        self.file_management_message_checks = [
            self.is_file_purge_command,
            self.is_file_delete_command,
            self.is_file_binary_response,
            self.is_file_upload_initiate,
            self.is_file_upload_abort,
            self.is_file_list,
            self.is_file_url_initiate,
            self.is_file_url_abort,
        ]

        self.firmware_update_message_checks = [
            self.is_firmware_install,
            self.is_firmware_abort,
        ]

    def _form_topic(self, message_type: str) -> str:
        return (
            self.PLATFORM_TO_DEVICE
            + self.key
            + self.CHANNEL_DELIMITER
            + message_type
        )

    def get_inbound_topics(self) -> List[str]:
        """
        Return list of inbound topics for device.

        :returns: List of topics to subscribe to
        :rtype: List[str]
        """
        return self.inbound_topics

    def is_time_response(self, message: Message) -> bool:
        """
        Check if message is response to time request.

        :param message: The message received
        :type message: Message
        :returns: is_time_response
        :rtype: bool
        """
        is_time_response = message.topic == self.time_topic
        self.logger.debug(
            f"{message.topic} is time response: {is_time_response}"
        )
        return is_time_response

    def is_feed_values(self, message: Message) -> bool:
        """
        Check if message is for incoming feed values.

        :param message: The message received
        :type message: Message
        :returns: is_feed_values
        :rtype: bool
        """
        is_feed_values = message.topic == self.feed_values_topic
        self.logger.debug(f"{message.topic} is feed values: {is_feed_values}")
        return is_feed_values

    def is_parameters(self, message: Message) -> bool:
        """
        Check if message is for updating device parameters.

        :param message: The message received
        :type message: Message
        :returns: is_parameters
        :rtype: bool
        """
        is_parameters = message.topic == self.parameters_topic
        self.logger.debug(
            f"{message.topic} is parameters message: " f"{is_parameters}"
        )
        return is_parameters

    def is_file_management_message(self, message: Message) -> bool:
        """
        Check if message is any kind of file management related message.

        :param message: The message received
        :type message: Message
        :returns: is_file_management_message
        :rtype: bool
        """
        return any(
            is_message(message)
            for is_message in self.file_management_message_checks
        )

    def is_firmware_message(self, message: Message) -> bool:
        """
        Check if message is any kind of firmware related message.

        :param message: The message received
        :type message: Message
        :returns: is_firmware_message
        :rtype: bool
        """
        return any(
            is_message(message)
            for is_message in self.firmware_update_message_checks
        )

    def is_firmware_install(self, message: Message) -> bool:
        """
        Check if message is firmware update install command.

        :param message: The message received
        :type message: Message
        :returns: firmware_update_install
        :rtype: bool
        """
        firmware_update_install = message.topic == self.firmware_install_topic
        self.logger.debug(
            f"{message.topic} is firmware install: {firmware_update_install}"
        )
        return firmware_update_install

    def is_firmware_abort(self, message: Message) -> bool:
        """
        Check if message is firmware update command.

        :param message: The message received
        :type message: Message
        :returns: firmware_update_abort
        :rtype: bool
        """
        firmware_update_abort = message.topic == self.firmware_abort_topic
        self.logger.debug(
            f"{message.topic} is firmware abort: {firmware_update_abort}"
        )
        return firmware_update_abort

    def is_file_binary_response(self, message: Message) -> bool:
        """
        Check if message is file binary message.

        :param message: The message received
        :type message: Message
        :returns: file_binary
        :rtype: bool
        """
        file_binary = message.topic == self.file_binary_topic
        self.logger.debug(f"{message.topic} is file binary: {file_binary}")
        return file_binary

    def is_file_delete_command(self, message: Message) -> bool:
        """
        Check if message if file delete command.

        :param message: The message received
        :type message: Message
        :returns: file_delete_command
        :rtype: bool
        """
        file_delete_command = message.topic == self.file_delete_topic
        self.logger.debug(
            f"{message.topic} is file delete: {file_delete_command}"
        )
        return file_delete_command

    def is_file_purge_command(self, message: Message) -> bool:
        """
        Check if message if file purge command.

        :param message: The message received
        :type message: Message
        :returns: file_purge_command
        :rtype: bool
        """
        file_purge_command = message.topic == self.file_purge_topic
        self.logger.debug(
            f"{message.topic} is file purge: {file_purge_command}"
        )
        return file_purge_command

    def is_file_list(self, message: Message) -> bool:
        """
        Check if message is file list request message.

        :param message: The message received
        :type message: Message
        :returns: file_list
        :rtype: bool
        """
        file_list = message.topic == self.file_list
        self.logger.debug(f"{message.topic} is file list request: {file_list}")
        return file_list

    def is_file_upload_initiate(self, message: Message) -> bool:
        """
        Check if message is file upload command.

        :param message: The message received
        :type message: Message
        :returns: file_upload_initiate
        :rtype: bool
        """
        file_upload_initiate = message.topic == self.file_upload_initiate_topic
        self.logger.debug(
            f"{message.topic} is file upload initiate: {file_upload_initiate}"
        )
        return file_upload_initiate

    def is_file_upload_abort(self, message: Message) -> bool:
        """
        Check if message is file upload command.

        :param message: The message received
        :type message: Message
        :returns: file_upload_abort_command
        :rtype: bool
        """
        file_upload_abort_command = (
            message.topic == self.file_upload_abort_topic
        )
        self.logger.debug(
            f"{message.topic} is file upload abort: {file_upload_abort_command}"
        )
        return file_upload_abort_command

    def is_file_url_initiate(self, message: Message) -> bool:
        """
        Check if message is file URL download command.

        :param message: The message received
        :type message: Message
        :returns: file_url_download_init
        :rtype: bool
        """
        file_url_download_init = message.topic == self.file_url_initiate_topic
        self.logger.debug(
            f"{message.topic} is file URL download: {file_url_download_init}"
        )
        return file_url_download_init

    def is_file_url_abort(self, message: Message) -> bool:
        """
        Check if message is file URL download command.

        :param message: The message received
        :type message: Message
        :returns: file_url_download_abort
        :rtype: bool
        """
        file_url_download_abort = message.topic == self.file_url_abort_topic
        self.logger.debug(
            f"{message.topic} is file URL abort: {file_url_download_abort}"
        )
        return file_url_download_abort

    def parse_time_response(self, message: Message) -> int:
        """
        Parse the message into an UTC timestamp.

        :param message: The message received
        :type message: Message
        :returns: timestamp
        :rtype: int
        """
        self.logger.debug(f"{message}")
        payload = json.loads(message.payload.decode("utf-8"))  # type: ignore

        timestamp = payload

        self.logger.debug(f"received timestamp: {timestamp}")
        return timestamp

    def parse_firmware_install(self, message: Message) -> str:
        """
        Return file name from message.

        :param message: The message received
        :type message: Message
        :returns: file_name
        :rtype: str
        """
        try:
            payload = json.loads(
                message.payload.decode("utf-8")  # type: ignore
            )
            file_name = payload
        except Exception:
            self.logger.warning(
                f"Received invalid firmware install message: {message}"
            )
            file_name = ""

        self.logger.debug(f"File name: {file_name}")
        return file_name

    def parse_file_binary(self, message: Message) -> FileTransferPackage:
        """
        Parse the message into a file transfer package.

        :param message: The message received
        :type message: Message
        :returns: file_transfer_package
        :rtype: FileTransferPackage
        """
        if not isinstance(message.payload, bytes):
            self.logger.warning(
                f"Expected payload in bytes, got {type(message.payload)}"
            )
            self.logger.info("Treating as invalid package")
            file_transfer_package = FileTransferPackage(b"", b"", b"")
        else:
            try:
                if len(message.payload) < 65:
                    raise ValueError(
                        "Received file transfer package too small"
                    )
                previous_hash = message.payload[:32]
                data = message.payload[32 : len(message.payload) - 32]
                current_hash = message.payload[-32:]

                file_transfer_package = FileTransferPackage(
                    previous_hash, data, current_hash
                )
                self.logger.debug(
                    "Received file transfer package: "
                    f"FileTransferPackage(previous_hash={previous_hash!r},"
                    f" data={len(data)} bytes, current_hash={current_hash!r}"
                )
            except Exception:
                self.logger.warning("Received malformed file chunk!")
                file_transfer_package = FileTransferPackage(b"", b"", b"")
        return file_transfer_package

    def parse_file_delete_command(self, message: Message) -> List[str]:
        """
        Parse the message into a list of file names.

        :param message: The message received
        :type message: Message
        :returns: file_name
        :rtype: List[str]
        """
        self.logger.debug(f"{message}")
        try:
            payload = json.loads(
                message.payload.decode("utf-8")  # type: ignore
            )
            self.logger.debug(f"file names: {payload}")
            return payload
        except Exception:
            self.logger.warning(
                f"Failed to get file name from message {message}"
            )
            return []

    def parse_file_url(self, message: Message) -> str:
        """
        Parse the message into a URL string.

        :param message: The message received
        :type message: Message
        :returns: file_url
        :rtype: str
        """
        self.logger.debug(f"{message}")
        try:
            payload = json.loads(
                message.payload.decode("utf-8")  # type: ignore
            )
            return payload
        except Exception:
            self.logger.warning(
                f"Failed to get file URL from message {message}"
            )
            return ""

    def parse_file_initiate(self, message: Message) -> Tuple[str, int, str]:
        """
        Return file name, file size and file hash from message.

        :param message: The message received
        :type message: Message
        :returns: (name, size, hash)
        :rtype: Tuple[str, int, str]
        """
        self.logger.debug(f"{message}")
        try:
            payload = json.loads(
                message.payload.decode("utf-8")  # type: ignore
            )
            self.logger.debug(
                f'name={payload["name"]}, '
                f'size={payload["size"]}, '
                f'hash={payload["hash"]}'
            )
            return (
                payload["name"],
                payload["size"],
                payload["hash"],
            )
        except Exception:
            self.logger.warning(
                f"Received invalid file upload initiate message: {message}"
            )
            return "", 0, ""

    def parse_parameters(
        self, message: Message
    ) -> Dict[str, Union[bool, int, float, str]]:
        """
        Parse the incoming parameters message.

        :param message: The message received
        :type message: Message
        :returns: parameters
        :rtype: Dict[str, Union[bool, int, float, str]]
        """
        self.logger.debug(f"{message}")
        try:
            parameters = json.loads(message.payload)
            return parameters
        except Exception as e:
            self.logger.exception(f"Failed to parse parameters message: {e}")
            return {}

    def parse_feed_values(
        self, message: Message
    ) -> List[Dict[str, Union[bool, int, float, str]]]:
        """
        Parse the incoming feed values message.

        :param message: The message received
        :type message: Message
        :returns: feed_values
        :rtype: List[Dict[str, Union[bool, int, float, str]]]
        """
        self.logger.debug(f"{message}")
        try:
            feed_values = json.loads(message.payload)
            return feed_values
        except Exception as e:
            self.logger.exception(f"Failed to parse feed values message: {e}")
            return []



"""Create messages from data that conform to device's specified protocol."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from abc import ABC
from abc import abstractmethod
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from wolk.model.data_type import DataType
from wolk.model.feed_type import FeedType
from wolk.model.file_management_status import FileManagementStatus
from wolk.model.firmware_update_status import FirmwareUpdateStatus
from wolk.model.message import Message
from wolk.model.unit import Unit

OutgoingDataTypes = Union[bool, int, float, str]
Reading = Tuple[str, OutgoingDataTypes]


class MessageFactory(ABC):
    """Serialize messages to be sent to WolkAbout IoT Platform."""

    @abstractmethod
    def make_from_feed_value(
        self,
        reading: Union[Reading, List[Reading]],
        timestamp: Optional[int],
    ) -> Message:
        """
        Serialize feed value data.

        :param reading: Feed value data as (reference, value) or list of tuple
        :type reading: Union[Reading, List[Reading]]
        :param timestamp: Unix timestamp in ms. Default to current time if None
        :returns: message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_from_feed_values_collected(
        self, collected_readings: Dict[int, Dict[str, OutgoingDataTypes]]
    ) -> Message:
        """
        Serialize feed values collected and organized by timestamp.

        :param collected_readings: Feed values, organized by timestamp, and then by reference.
        :type collected_readings: Dict[int, Dict[str, OutgoingDataTypes]]
        :return: The message containing all the data
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_pull_feed_values(self) -> Message:
        """
        Serialize message requesting any pending inbound feed values.

        :returns: message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_time_request(self) -> Message:
        """
        Serialize message requesting platform timestamp.

        :returns: message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_from_parameters(
        self, parameters: Dict[str, Union[bool, int, float, str]]
    ) -> Message:
        """
        Serialize device parameters to be sent to the Platform.

        :param parameters: Device parameters
        :type parameters: Dict[str, Union[bool, int, float, str]]
        :returns: message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_pull_parameters(self) -> Message:
        """
        Serialize request to pull device parameters from the Platform.

        :returns: message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_feed_registration(
        self,
        name: str,
        reference: str,
        feed_type: FeedType,
        unit: Union[Unit, str],
    ) -> Message:
        """
        Serialize request to register a feed for the device on the Platform.

        :param name: Feed name
        :type name: str
        :param reference: Unique identifier
        :type reference: str
        :param feed_type: Is the feed one or two-way communication
        :type feed_type: FeedType
        :param unit: Unit used to measure this feed
        :type unit: Union[Unit, str]
        :returns: message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_feed_removal(self, reference: str) -> Message:
        """
        Serialize request to remove a feed from the device on the Platform.

        :param reference: Unique identifier
        :type reference: str
        :returns: message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_attribute_registration(
        self, name: str, data_type: DataType, value: str
    ) -> Message:
        """
        Serialize request to register an attribute for the device.

        :param name: Unique identifier
        :type name: str
        :param data_type: Type of data this attribute holds
        :type data_type: DataType
        :param value: Value of the attribute
        :type value: str
        :returns: message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_from_firmware_update_status(
        self, firmware_update_status: FirmwareUpdateStatus
    ) -> Message:
        """
        Report the current status of the firmware update process.

        :param firmware_update_status: Status of the firmware update process
        :type firmware_update_status: FirmwareUpdateStatus
        :returns: message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_from_package_request(
        self, file_name: str, chunk_index: int
    ) -> Message:
        """
        Request a package of the file from WolkAbout IoT Platform.

        :param file_name: Name of the file that contains the requested package
        :type file_name: str
        :param chunk_index: Index of the requested package
        :type chunk_index: int
        :returns: message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_from_file_list(
        self, file_list: List[Dict[str, Union[str, int]]]
    ) -> Message:
        """
        Serialize list of files present on device.

        :param file_list: Files present on device
        :type file_list: List[Dict[str, Union[str, int]]]
        :returns: message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_from_file_management_status(
        self, status: FileManagementStatus, file_name: str
    ) -> Message:
        """
        Serialize device's current file management status.

        :param status: Current file management status
        :type status: FileManagementStatus
        :param file_name: Name of file being transferred
        :type file_name: str
        :returns: message
        :rtype: Message
        """

    @abstractmethod
    def make_from_file_url_status(
        self,
        file_url: str,
        status: FileManagementStatus,
        file_name: Optional[str] = None,
    ) -> Message:
        """
        Serialize device's current file URL download status.

        :param file_url: URL from where the file is to be downloaded
        :type file_url: str
        :param status: Current file management status
        :type status: FileManagementStatus
        :param file_name: Only present when download of file is completed
        :type file_name: Optional[str]
        """
        raise NotImplementedError()



"""Tests for LoggerFactory."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import logging
import sys
import unittest

sys.path.append("..")  # noqa

from wolk import logger_factory


class TestLoggerFactory(unittest.TestCase):
    """Tests for LoggerFactory class."""

    def test_get_logger_with_level(self):
        """Test getting defualt logger with specified level."""
        logger = logger_factory.logger_factory.get_logger(
            "name", logging.NOTSET
        )
        self.assertEqual(logging.NOTSET, logger.level)

    def test_logging_config_logger_with_log_file(self):
        """Test getting logger with log file."""
        test_log = "test.log"
        logger_factory.logging_config("info", test_log)
        logger = logger_factory.logger_factory.get_logger(
            "test", logging.NOTSET
        )
        self.assertIsInstance(logger.handlers[1], logging.FileHandler)

        logger.handlers[1].close()
        logger.removeHandler(logger.handlers[1])
        logger_factory.logger_factory.log_file = None
        import os

        os.remove(test_log)

    def test_logging_config_logger_only_log_file(self):
        """Test getting logger only with log file."""
        test_log = "test.log"
        logger_factory.logging_config("info", test_log)
        logger_factory.logger_factory.console = False
        logger = logger_factory.logger_factory.get_logger("test")
        self.assertIsInstance(logger.handlers[0], logging.FileHandler)

        logger_factory.logger_factory.console = True
        logger.handlers[0].close()
        logger.removeHandler(logger.handlers[0])
        logger_factory.logger_factory.log_file = None
        import os

        os.remove(test_log)

    def test_logging_config_debug(self):
        """Test setting log level to debug."""
        logger_factory.logging_config("debug")
        self.assertEqual(logging.DEBUG, logger_factory.logger_factory.level)

    def test_logging_config_info_then_invalid(self):
        """Test setting log level to info and then to invalid value."""
        logger_factory.logging_config("info")
        logger_factory.logging_config("tests")
        self.assertEqual(logging.INFO, logger_factory.logger_factory.level)



"""Module responsible for handling files and file transfer."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from abc import ABC
from abc import abstractmethod
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from wolk.model.file_management_status import FileManagementStatus
from wolk.model.file_transfer_package import FileTransferPackage


class FileManagement(ABC):
    """
    File transfer manager.

    Enables device to transfer files from WolkAbout IoT Platform
     package by package or/and URL download.
    """

    @abstractmethod
    def __init__(
        self,
        status_callback: Callable[[str, FileManagementStatus], None],
        packet_request_callback: Callable[[str, int, int], None],
        url_status_callback: Callable[
            [str, FileManagementStatus, Optional[str]], None
        ],
    ) -> None:
        """
        Enable file management for device.

        :param status_callback: Reporting current file transfer status
        :type status_callback: Callable[[FileManagementStatus], None]
        :param packet_request_callback: Request file packet from Platform
        :type packet_request_callback: Callable[[str, int, int], None]
        :param url_status_callback: Report file url download status
        :type url_status_callback: Callable[[str, FileManagementStatus, Optional[str]], None]
        """
        raise NotImplementedError()

    @abstractmethod
    def get_preffered_package_size(self) -> int:
        """
        Return preffered package size for file transfer.

        :returns: preferred_package_size
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def supports_url_download(self) -> bool:
        """
        Return if the file management module supports URL download.

        :returns: supports_url_download
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def configure(
        self,
        file_directory: str,
        preferred_package_size: int = 0,
    ) -> None:
        """
        Configure options for file management module.

        :param file_directory: Path to where files are stored
        :type file_directory: str
        :param preferred_package_size: Size in kilobytes, 0 means no limit
        :type preferred_package_size: int
        """
        raise NotImplementedError()

    @abstractmethod
    def set_custom_url_downloader(
        self, downloader: Callable[[str, str], bool]
    ) -> None:
        """
        Set the URL file downloader to a custom implementation.

        :param downloader: Function that will download the file from the URL
        :type downloader: Callable[[str, str], bool]
        """
        raise NotImplementedError()

    @abstractmethod
    def handle_upload_initiation(
        self, file_name: str, file_size: int, file_hash: str
    ) -> None:
        """
        Start making package requests and set status to file transfer.

        :param file_name: File name
        :type file_name: str
        :param file_size: Size in bytes
        :type file_size: int
        :param file_hash: base64 encoded sha256 hash of file
        :type file_hash: str
        """
        raise NotImplementedError()

    @abstractmethod
    def handle_file_upload_abort(self) -> None:
        """Abort file upload and revert to idle status."""
        raise NotImplementedError()

    @abstractmethod
    def handle_file_binary_response(
        self, package: FileTransferPackage
    ) -> None:
        """
        Validate received package and store or use callback to request again.

        :param package: Package of file being transfered.
        :type package: FileTransferPackage
        """
        raise NotImplementedError()

    @abstractmethod
    def handle_file_url_download_initiation(self, file_url: str) -> None:
        """
        Start file transfer from specified URL.

        :param file_url: URL from where to download file
        :type file_url: str
        """
        raise NotImplementedError()

    @abstractmethod
    def handle_file_url_download_abort(self) -> None:
        """Abort file URL download."""
        raise NotImplementedError()

    @abstractmethod
    def get_file_list(self) -> List[Dict[str, Union[str, int]]]:
        """
        Return list of files present on device.

        Each list item is a dictionary that contains the name of the file,
        its size in bytes, and a MD5 checksum of the file.

        :returns: file_list
        :rtype: List[Dict[str, Union[str, int]]]

        """
        raise NotImplementedError()

    @abstractmethod
    def get_file_path(self, file_name: str) -> Optional[str]:
        """
        Return path to file if it exists.

        :param file_name: File for which to get path
        :type file_name: str
        :returns: file_path
        :rtype: Optional[str]
        """
        raise NotImplementedError()

    @abstractmethod
    def handle_file_delete(self, file_names: List[str]) -> None:
        """
        Delete files from device.

        :param file_names: Files to be deleted
        :type file_names: List[str]
        """
        raise NotImplementedError()

    @abstractmethod
    def handle_file_purge(self) -> None:
        """Delete all files from device."""
        raise NotImplementedError()



"""Enumeration of defined file management errors."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from enum import Enum


class FileManagementErrorType(Enum):
    """Enumeration of available file management errors."""

    UNKNOWN = "UNKNOWN"
    TRANSFER_PROTOCOL_DISABLED = "TRANSFER_PROTOCOL_DISABLED"
    UNSUPPORTED_FILE_SIZE = "UNSUPPORTED_FILE_SIZE"
    MALFORMED_URL = "MALFORMED_URL"
    FILE_HASH_MISMATCH = "FILE_HASH_MISMATCH"
    FILE_SYSTEM_ERROR = "FILE_SYSTEM_ERROR"
    RETRY_COUNT_EXCEEDED = "RETRY_COUNT_EXCEEDED"



"""Service for exchanging data with WolkAbout IoT Platform."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from abc import ABC
from abc import abstractmethod
from typing import Callable

from wolk.model.message import Message


class ConnectivityService(ABC):
    """Responsible for exchanging data with WolkAbout IoT Platform."""

    @abstractmethod
    def connect(self) -> bool:
        """Connect to WolkAbout IoT Platform."""
        raise NotImplementedError()

    @abstractmethod
    def disconnect(self) -> None:
        """Disconnect from WolkAbout IoT Platform."""
        raise NotImplementedError()

    @abstractmethod
    def is_connected(self) -> bool:
        """
        Return current connection state.

        :returns: connected
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def publish(self, message: Message) -> bool:
        """
        Publish a message to WolkAbout IoT Platform.

        :param outbound_message: Message to send
        :type outbound_message: Message
        :returns: success
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def set_inbound_message_listener(
        self, listener: Callable[[Message], None]
    ) -> None:
        """
        Set a callback method to handle inbound messages.

        :param listener: Method hat handles inbound messages
        :type listener: Callable[[Message], None]
        """
        raise NotImplementedError()



"""Process messages received from WolkAbout IoT Platform."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from abc import ABC
from abc import abstractmethod
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union

from wolk.model.file_transfer_package import FileTransferPackage
from wolk.model.message import Message


class MessageDeserializer(ABC):
    """Deserialize messages received from the platform."""

    @abstractmethod
    def get_inbound_topics(self) -> List[str]:
        """
        Return list of inbound topics for device.

        :returns: List of topics to subscribe to
        :rtype: List[str]
        """
        raise NotImplementedError()

    @abstractmethod
    def is_time_response(self, message: Message) -> bool:
        """
        Check if message is response to time request.

        :param message: The message received
        :type message: Message
        :returns: is_time_response
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_file_management_message(self, message: Message) -> bool:
        """
        Check if message is any kind of file management related message.

        :param message: The message received
        :type message: Message
        :returns: is_file_management_message
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_firmware_message(self, message: Message) -> bool:
        """
        Check if message is any kind of firmware related message.

        :param message: The message received
        :type message: Message
        :returns: is_firmware_message
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_feed_values(self, message: Message) -> bool:
        """
        Check if message is for incoming feed values.

        :param message: The message received
        :type message: Message
        :returns: is_feed_values
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_parameters(self, message: Message) -> bool:
        """
        Check if message is for updating device parameters.

        :param message: The message received
        :type message: Message
        :returns: is_parameters
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_firmware_install(self, message: Message) -> bool:
        """
        Check if message is firmware update install command.

        :param message: The message received
        :type message: Message
        :returns: firmware_update_install_command
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_firmware_abort(self, message: Message) -> bool:
        """
        Check if message is firmware update command.

        :param message: The message received
        :type message: Message
        :returns: firmware_update_abort_command
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_file_binary_response(self, message: Message) -> bool:
        """
        Check if message is file binary message.

        :param message: The message received
        :type message: Message
        :returns: file_binary_response
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_file_delete_command(self, message: Message) -> bool:
        """
        Check if message if file delete command.

        :param message: The message received
        :type message: Message
        :returns: file_delete_command
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_file_purge_command(self, message: Message) -> bool:
        """
        Check if message if file purge command.

        :param message: The message received
        :type message: Message
        :returns: file_purge_command
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_file_list(self, message: Message) -> bool:
        """
        Check if message is file list request message.

        :param message: The message received
        :type message: Message
        :returns: file_list
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_file_upload_initiate(self, message: Message) -> bool:
        """
        Check if message is file upload command.

        :param message: The message received
        :type message: Message
        :returns: file_upload_initiate_command
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_file_upload_abort(self, message: Message) -> bool:
        """
        Check if message is file upload command.

        :param message: The message received
        :type message: Message
        :returns: file_upload_abort_command
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_file_url_initiate(self, message: Message) -> bool:
        """
        Check if message is file URL download command.

        :param message: The message received
        :type message: Message
        :returns: file_url_download_initiate
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def is_file_url_abort(self, message: Message) -> bool:
        """
        Check if message is file URL download command.

        :param message: The message received
        :type message: Message
        :returns: file_url_download_abort
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def parse_time_response(self, message: Message) -> int:
        """
        Parse the message into an UTC timestamp.

        :param message: The message received
        :type message: Message
        :returns: timestamp
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def parse_firmware_install(self, message: Message) -> str:
        """
        Return file name from message.

        :param message: The message received
        :type message: Message
        :returns: file_name
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def parse_file_initiate(self, message: Message) -> Tuple[str, int, str]:
        """
        Return file name, file size and file hash from message.

        :param message: The message received
        :type message: Message
        :returns: (file_name, file_size, file_hash)
        :rtype: Tuple[str, int, str]
        """
        raise NotImplementedError()

    @abstractmethod
    def parse_file_url(self, message: Message) -> str:
        """
        Parse the message into a URL string.

        :param message: The message received
        :type message: Message
        :returns file_url:
        :rtype: str
        """
        raise NotImplementedError()

    @abstractmethod
    def parse_file_binary(self, message: Message) -> FileTransferPackage:
        """
        Parse the message into a file transfer package.

        :param message: The message received
        :type message: Message
        :returns: file_transfer_package
        :rtype: FileTransferPackage
        """
        raise NotImplementedError()

    @abstractmethod
    def parse_file_delete_command(self, message: Message) -> List[str]:
        """
        Parse the message into a list of file names.

        :param message: The message received
        :type message: Message
        :returns: file_name
        :rtype: List[str]
        """
        raise NotImplementedError()

    @abstractmethod
    def parse_parameters(
        self, message: Message
    ) -> Dict[str, Union[bool, int, float, str]]:
        """
        Parse the incoming parameters message.

        :param message: The message received
        :type message: Message
        :returns: parameters
        :rtype: Dict[str, Union[bool, int, float, str]]
        """
        raise NotImplementedError()

    @abstractmethod
    def parse_feed_values(
        self, message: Message
    ) -> List[Dict[str, Union[bool, int, float, str]]]:
        """
        Parse the incoming feed values message.

        :param message: The message received
        :type message: Message
        :returns: feed_values
        :rtype: List[Dict[str, Union[bool, int, float, str]]]
        """
        raise NotImplementedError()



"""LoggerFactory Module."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import logging
from typing import List
from typing import Optional


class LoggerFactory:
    """Factory for issuing ready to use loggers in other modules."""

    def __init__(self, level=logging.INFO, console=True, log_file=None):  # type: ignore
        """
        Create a factory that will give loggers through calls to get_logger().

        :param level: Set the desired logging level
        :type level: int or None
        :param console: Should the log messages be outputted to the console
        :type console: bool or None
        :param log_file: Name of the log file to output to
        :type log_file: str or None
        """
        self.level = level
        self.device_key = None
        self.console = console
        self.log_file = log_file
        self.loggers: List[logging.Logger] = []

    def set_device_key(self, device_key: str) -> None:
        """
        Set device key.

        :param device_key: Device key
        :type device_key: str
        """
        self.device_key = device_key

    def get_logger(
        self, name: str, level: Optional[int] = None
    ) -> logging.Logger:
        """
        Return a ready to use logger instance.

        :param name: Name of the logger
        :type name: str
        :param level: Override the log level
        :type level: int or None

        :returns: Logger instance
        :rtype: logger
        """
        logger = logging.getLogger(name)

        if level is not None:
            logger.setLevel(level)
        else:
            logger.setLevel(self.level)

        if self.device_key is not None:
            formatter = logging.Formatter(
                "%(asctime)s - '"
                + str(self.device_key)
                + "' - %(levelname)s [%(filename)s:%(lineno)s"
                + " - %(funcName)s()] - %(message)s"
            )
        else:
            formatter = logging.Formatter(
                "%(asctime)s - %(levelname)s [%(filename)s:%(lineno)s"
                + " - %(funcName)s()] - %(message)s"
            )
        if self.console:

            console_handler = logging.StreamHandler()
            if level is not None:
                console_handler.setLevel(level)
            else:
                console_handler.setLevel(self.level)

            console_handler.setFormatter(formatter)
            logger.addHandler(console_handler)

        if self.log_file is not None:
            file_handler = logging.FileHandler(self.log_file)
            if level is not None:
                file_handler.setLevel(level)
            else:
                file_handler.setLevel(self.level)
            file_handler.setFormatter(formatter)
            logger.addHandler(file_handler)

        self.loggers.append(logger)
        return logger


# Logging levels available: NOTSET, INFO, DEBUG
logger_factory = LoggerFactory(level=logging.INFO)

LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
    "notset": logging.NOTSET,
}


def logging_config(level: str, log_file: Optional[str] = None) -> None:
    """
    Set desired log level and designate a log file.

    :param level: Available levels : debug, info, notset
    :type level: str
    :param log_file: path to log file
    :type log_file: str or None
    """
    if log_file is not None:
        logger_factory.log_file = log_file

    if level not in LEVELS:
        print(f"Invalid level '{level}'")
        return

    if LEVELS[level] == logger_factory.level:
        return

    logger_factory.level = LEVELS[level]

    for logger in logger_factory.loggers:
        logger.setLevel(logger_factory.level)
        for handler in logger.handlers:
            handler.setLevel(logger_factory.level)



"""Enumeration of available units on the Platform."""
#   Copyright 2021 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from enum import Enum


class Unit(Enum):
    """
    Enumeration of default available units on the Platform.

    Units are grouped by reading type, along with a comment
    that indicates what data type that unit expects.

    """

    # GENERIC
    TEXT = "TEXT"  # STRING
    NUMERIC = "NUMERIC"  # NUMERIC
    BOOLEAN = "BOOLEAN"  # BOOLEAN
    PERCENT = "PERCENT"  # PERCENT
    DECIBEL = "DECIBEL"  # NUMERIC
    LOCATION = "LOCATION"  # LOCATION

    # ACCELERATION
    METRES_PER_SQUARE_SECOND = "METRES_PER_SQUARE_SECOND"  # NUMERIC
    G = "G"  # NUMERIC

    # AMOUNT_OF_SUBSTANCE
    MOLE = "MOLE"  # NUMERIC
    ATOM = "ATOM"  # NUMERIC

    # ANGLE
    RADIAN = "RADIAN"  # NUMERIC
    REVOLUTION = "REVOLUTION"  # NUMERIC
    DEGREE_ANGLE = "DEGREE_ANGLE"  # NUMERIC
    MINUTE_ANGLE = "MINUTE_ANGLE"  # NUMERIC
    SECOND_ANGLE = "SECOND_ANGLE"  # NUMERIC
    CENTIRADIAN = "CENTIRADIAN"  # NUMERIC
    GRADE = "GRADE"  # NUMERIC

    # AREA
    SQUARE_METRE = "SQUARE_METRE"  # NUMERIC
    ARE = "ARE"  # NUMERIC
    HECTARE = "HECTARE"  # NUMERIC

    # CATALYTIC_ACTIVITY
    KATAL = "KATAL"  # NUMERIC

    # DATA_AMOUNT
    BIT = "BIT"  # NUMERIC
    BYTE = "BYTE"  # NUMERIC

    # DURATION
    SECOND = "SECOND"  # NUMERIC
    MINUTE = "MINUTE"  # NUMERIC
    HOUR = "HOUR"  # NUMERIC
    DAY = "DAY"  # NUMERIC
    WEEK = "WEEK"  # NUMERIC
    YEAR = "YEAR"  # NUMERIC
    MONTH = "MONTH"  # NUMERIC
    DAY_SIDEREAL = "DAY_SIDEREAL"  # NUMERIC
    YEAR_SIDEREAL = "YEAR_SIDEREAL"  # NUMERIC
    YEAR_CALENDAR = "YEAR_CALENDAR"  # NUMERIC

    # DYNAMIC_VISCOSITY
    POISE = "POISE"  # NUMERIC

    # ELECTRIC_CAPACITANCE
    FARAD = "FARAD"  # NUMERIC

    # ELECTRIC_CHARGE
    COULOMB = "COULOMB"  # NUMERIC
    E = "E"  # NUMERIC
    FARADAY = "FARADAY"  # NUMERIC
    FRANKLIN = "FRANKLIN"  # NUMERIC

    # ELECTRIC_CONDUCTANCE
    SIEMENS = "SIEMENS"  # NUMERIC

    # ELECTRIC_CURRENT
    AMPERE = "AMPERE"  # NUMERIC
    GILBERT = "GILBERT"  # NUMERIC

    # ELECTRIC_INDUCTANCE
    HENRY = "HENRY"  # NUMERIC

    # ELECTRIC_POTENTIAL
    VOLT = "VOLT"  # NUMERIC
    CENTIVOLT = "CENTIVOLT"  # NUMERIC
    MILLIVOLT = "MILLIVOLT"  # NUMERIC

    # ELECTRIC_RESISTANCE
    OHM = "OHM"  # NUMERIC

    # ENERGY
    JOULE = "JOULE"  # NUMERIC
    ERG = "ERG"  # NUMERIC
    ELECTRON_VOLT = "ELECTRON_VOLT"  # NUMERIC

    # FORCE
    NEWTON = "NEWTON"  # NUMERIC
    DYNE = "DYNE"  # NUMERIC
    KILOGRAM_FORCE = "KILOGRAM_FORCE"  # NUMERIC
    POUND_FORCE = "POUND_FORCE"  # NUMERIC

    # FREQUENCY
    HERTZ = "HERTZ"  # NUMERIC
    MEGAHERTZ = "MEGAHERTZ"  # NUMERIC
    GIGAHERTZ = "GIGAHERTZ"  # NUMERIC

    # ILLUMINANCE
    LUX = "LUX"  # NUMERIC
    LAMBERT = "LAMBERT"  # NUMERIC

    # KINEMATIC_VISCOSITY
    STOKE = "STOKE"  # NUMERIC

    # LENGTH
    METRE = "METRE"  # NUMERIC
    KILOMETRE = "KILOMETRE"  # NUMERIC
    CENTIMETRE = "CENTIMETRE"  # NUMERIC
    MILLIMETRE = "MILLIMETRE"  # NUMERIC
    FOOT = "FOOT"  # NUMERIC
    FOOT_SURVEY_US = "FOOT_SURVEY_US"  # NUMERIC
    YARD = "YARD"  # NUMERIC
    INCH = "INCH"  # NUMERIC
    MILE = "MILE"  # NUMERIC
    NAUTICAL_MILE = "NAUTICAL_MILE"  # NUMERIC
    ANGSTROM = "ANGSTROM"  # NUMERIC
    ASTRONOMICAL_UNIT = "ASTRONOMICAL_UNIT"  # NUMERIC
    LIGHT_YEAR = "LIGHT_YEAR"  # NUMERIC
    PARSEC = "PARSEC"  # NUMERIC
    POINT = "POINT"  # NUMERIC
    PIXEL = "PIXEL"  # NUMERIC

    # LUMINOUS_FLUX
    LUMEN = "LUMEN"  # NUMERIC

    # LUMINOUS_INTENSITY
    CANDELA = "CANDELA"  # NUMERIC

    # MAGNETIC_FLUX
    WEBER = "WEBER"  # NUMERIC
    MAXWELL = "MAXWELL"  # NUMERIC

    # MAGNETIC_FLUX_DENSITY
    TESLA = "TESLA"  # NUMERIC
    GAUSS = "GAUSS"  # NUMERIC

    # MASS
    KILOGRAM = "KILOGRAM"  # NUMERIC
    GRAM = "GRAM"  # NUMERIC
    ATOMIC_MASS = "ATOMIC_MASS"  # NUMERIC
    ELECTRON_MASS = "ELECTRON_MASS"  # NUMERIC
    POUND = "POUND"  # NUMERIC
    OUNCE = "OUNCE"  # NUMERIC
    TON_US = "TON_US"  # NUMERIC
    TON_UK = "TON_UK"  # NUMERIC
    METRIC_TON = "METRIC_TON"  # NUMERIC

    # POWER
    WATT = "WATT"  # NUMERIC
    HORSEPOWER = "HORSEPOWER"  # NUMERIC

    # PRESSURE
    PASCAL = "PASCAL"  # NUMERIC
    HECTOPASCAL = "HECTOPASCAL"  # NUMERIC
    ATMOSPHERE = "ATMOSPHERE"  # NUMERIC
    BAR = "BAR"  # NUMERIC
    MILLIBAR = "MILLIBAR"  # NUMERIC
    MILLIMETER_OF_MERCURY = "MILLIMETER_OF_MERCURY"  # NUMERIC
    INCH_OF_MERCURY = "INCH_OF_MERCURY"  # NUMERIC

    # RADIATION_DOSE_ABSORBED
    GRAY = "GRAY"  # NUMERIC
    RAD = "RAD"  # NUMERIC

    # RADIATION_DOSE_EFFECTIVE
    SIEVERT = "SIEVERT"  # NUMERIC
    REM = "REM"  # NUMERIC

    # RADIOACTIVE_ACTIVITY
    BECQUEREL = "BECQUEREL"  # NUMERIC
    CURIE = "CURIE"  # NUMERIC
    RUTHERFORD = "RUTHERFORD"  # NUMERIC

    # ROENTGEN
    ROENTGEN = "ROENTGEN"  # NUMERIC

    # SOLID_ANGLE
    STERADIAN = "STERADIAN"  # NUMERIC
    SPHERE = "SPHERE"  # NUMERIC

    # TEMPERATURE
    KELVIN = "KELVIN"  # NUMERIC
    CELSIUS = "CELSIUS"  # NUMERIC
    RANKINE = "RANKINE"  # NUMERIC
    FAHRENHEIT = "FAHRENHEIT"  # NUMERIC

    # VELOCITY
    METRES_PER_SECOND = "METRES_PER_SECOND"  # NUMERIC
    MILES_PER_HOUR = "MILES_PER_HOUR"  # NUMERIC
    KILOMETRES_PER_HOUR = "KILOMETRES_PER_HOUR"  # NUMERIC
    KNOT = "KNOT"  # NUMERIC
    MACH = "MACH"  # NUMERIC
    C = "C"  # NUMERIC

    # VOLUME
    CUBIC_METRE = "CUBIC_METRE"  # NUMERIC
    LITRE = "LITRE"  # NUMERIC
    DECILITRE = "DECILITRE"  # NUMERIC
    MILLILITRE = "MILLILITRE"  # NUMERIC
    CUBIC_INCH = "CUBIC_INCH"  # NUMERIC
    GALLON_DRY_US = "GALLON_DRY_US"  # NUMERIC
    GALLON_UK = "GALLON_UK"  # NUMERIC
    OUNCE_LIQUID_UK = "OUNCE_LIQUID_UK"  # NUMERIC



"""Store messages before sending them to WolkAbout IoT Platform."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from abc import ABC
from abc import abstractmethod
from typing import Optional

from wolk.model.message import Message


class MessageQueue(ABC):
    """Store messages on device before publishing to WolkAbout IoT Platform."""

    @abstractmethod
    def put(self, message: Message) -> bool:
        """
        Place a message in storage.

        :param message: Message to be stored
        :type message: Message
        :returns: result
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def get(self) -> Optional[Message]:
        """
        Get a message from storage.

        :returns: message
        :rtype: Optional[Message]
        """
        raise NotImplementedError()

    @abstractmethod
    def peek(self) -> Optional[Message]:
        """
        Get a message without removing from storage.

        :returns: message
        :rtype: Optional[Message]
        """
        raise NotImplementedError()



"""Tests for InMemoryReadingsPersistence."""
#   Copyright 2023 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import unittest
from time import time

from wolk.in_memory_readings_persistence import InMemoryReadingsPersistence


class TestInMemoryReadingsPersistence(unittest.TestCase):
    """Tests for InMemoryReadingsPersistence class."""

    def test_adding_empty_list(self):
        """Test adding a single reading to the persistence."""
        self.persistence = InMemoryReadingsPersistence()
        self.assertFalse(self.persistence.store_reading([]))
        self.assertEqual(len(self.persistence.obtain_readings()), 0)

    def test_adding_single_int_reading(self):
        """Test adding a single int reading to the persistence."""
        self.persistence = InMemoryReadingsPersistence()
        self.assertTrue(self.persistence.store_reading(("T", 123)))
        self.assertEqual(len(self.persistence.obtain_readings()), 1)

    def test_adding_multiple_int_readings(self):
        """Test adding a list of int readings to the persistence."""
        self.persistence = InMemoryReadingsPersistence()
        self.assertTrue(
            self.persistence.store_reading(
                [("A", 123), ("B", 456), ("C", 789)]
            )
        )
        self.assertEqual(len(self.persistence.obtain_readings()), 1)

    def test_add_multiple_readings_at_different_times_and_clean(self):
        """Test adding a lot of different readings at different times."""
        self.persistence = InMemoryReadingsPersistence()
        self.assertTrue(
            self.persistence.store_reading(
                [("A", 1), ("B", 2)], round(time() * 1000) + 123511
            )
        )
        self.assertTrue(
            self.persistence.store_reading(
                [("B", 3), ("C", 4)], round(time() * 1000) + 871265
            )
        )
        self.assertTrue(
            self.persistence.store_reading(
                [("C", 5), ("A", 6)], round(time() * 1000) + 839681
            )
        )
        self.assertEqual(len(self.persistence.obtain_readings()), 3)

        self.assertTrue(self.persistence.clear_readings())
        self.assertEqual(len(self.persistence.obtain_readings()), 0)



"""Setup file."""
from os import path

import setuptools

from wolk import __version__

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name="wolk-connect",
    version=__version__,
    install_requires=["paho_mqtt>=1.5.1", "requests>=2.18.1"],
    include_package_data=True,
    license="Apache License 2.0",
    author="WolkAbout",
    author_email="info@wolkabout.com",
    description="Python 3 library for connecting to WolkAbout IoT Platform",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=["IoT", "WolkAbout", "Internet of Things"],
    url="https://github.com/Wolkabout/WolkConnect-Python",
    test_suite="test",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
        "Topic :: Internet",
        "Topic :: Communications",
        "Topic :: Software Development :: Embedded Systems",
    ],
)



"""Persistence for readings that uses local memory (volatile storage)."""
#   Copyright 2023 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from time import time
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from wolk import logger_factory
from wolk.interface.message_factory import OutgoingDataTypes
from wolk.interface.message_factory import Reading
from wolk.interface.readings_persistence import ReadingsPersistence


class InMemoryReadingsPersistence(ReadingsPersistence):
    """Persistence for readings that uses local memory (volatile storage)."""

    def __init__(self):  # type: ignore
        """Create the dictionary to store readings."""
        self.dict: Dict[int, Dict[str, OutgoingDataTypes]] = {}
        self.logger = logger_factory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )

    def store_reading(
        self,
        reading: Union[Reading, List[Reading]],
        timestamp: Optional[int] = None,
    ) -> bool:
        """
        Stores the readings from the user in the persistence.

        Groups up with other readings based on the timestamp.

        :param reading: The reading(s) that should be stored in persistence.
        :type reading: Union[Reading, Dict[str, Reading]]
        :param timestamp: The timestamp of when the readings were current.
        :type timestamp: Optional[int]
        :returns: Whether the data was successfully stored.
        :rtype: bool
        """
        self.logger.debug("Storing readings in persistence.")

        # Skip execution if the list is empty
        if isinstance(reading, List) and len(reading) == 0:
            self.logger.warning(
                "Skipping reading storing as the readings list is empty."
            )
            return False

        # Resolve the timestamp if it isn't passed
        timestamp = (
            timestamp if timestamp is not None else round(time() * 1000)
        )

        # If there isn't an entry for the timestamp, create one
        if timestamp not in self.dict.keys():
            self.dict[timestamp] = {}

        # Append all the values into the dictionary
        if isinstance(reading, List):
            for entry in reading:
                self.dict[timestamp][entry[0]] = entry[1]
        else:
            self.dict[timestamp][reading[0]] = reading[1]

        # List out all the members
        return True

    def obtain_readings(self) -> Dict[int, Dict[str, OutgoingDataTypes]]:
        """
        Return everything from the dictionary.

        :returns: The entire content of the reading dictionary.
        :rtype: Dict[int, Dict[str, Reading]]
        """
        return self.dict

    def obtain_readings_count(self) -> int:
        """
        Return count of readings in dictionary.

        :returns: Total count of readings.
        :rtype: int
        """
        count = 0
        for [_, readings] in self.dict.items():
            count += len(readings)
        return count

    def clear_readings(self) -> bool:
        """
        Clears the entire map of readings.

        :returns: Whether the clear was successful.
        :rtype: bool
        """
        self.dict.clear()
        return True



"""Tests for WolkAboutProtocolMessageFactory."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import json
import sys
import time
import unittest

sys.path.append("..")  # noqa

from wolk.model.message import Message
from wolk.model.data_type import DataType
from wolk.model.feed_type import FeedType
from wolk.model.unit import Unit
from wolk.model.file_management_status import FileManagementStatus
from wolk.model.file_management_error_type import FileManagementErrorType
from wolk.model.file_management_status_type import FileManagementStatusType
from wolk.model.firmware_update_status import FirmwareUpdateStatus
from wolk.model.firmware_update_status_type import FirmwareUpdateStatusType
from wolk.model.firmware_update_error_type import FirmwareUpdateErrorType
from wolk.wolkabout_protocol_message_factory import (
    WolkAboutProtocolMessageFactory as WAPMF,
)

unittest.util._MAX_LENGTH = 2000


class WolkAboutProtocolMessageFactoryTests(unittest.TestCase):
    """Tests for serializing messages using WolkAbout Protocol."""

    def setUp(self):
        """Set up commonly used values in tests."""
        self.device_key = "some_key"
        self.factory = WAPMF(self.device_key)

    def test_init(self):
        """Test that object is created with correct device key."""
        self.assertEqual(self.device_key, self.factory.device_key)

    def test_feed_value(self):
        """Test valid message for string feed."""
        reference = "SNL"
        value = "string"
        timestamp = round(time.time()) * 1000

        expected_topic = self.factory.common_topic + WAPMF.FEED_VALUES
        expected_payload = json.dumps(
            [{reference: value, "timestamp": timestamp}]
        )
        expected_message = Message(expected_topic, expected_payload)

        serialized_message = self.factory.make_from_feed_value(
            (reference, value), timestamp
        )

        self.assertEqual(expected_message, serialized_message)

    def test_feed_values(self):
        """Test valid message for two string feeds."""
        reference = "SNL"
        value = "string"
        reference_2 = "SNL"
        value_2 = "string"
        timestamp = round(time.time()) * 1000

        expected_topic = self.factory.common_topic + WAPMF.FEED_VALUES
        expected_payload = json.dumps(
            [{reference: value, "timestamp": timestamp}]
        )
        expected_message = Message(expected_topic, expected_payload)

        serialized_message = self.factory.make_from_feed_value(
            [(reference, value), (reference_2, value_2)], timestamp
        )

        self.assertEqual(expected_message, serialized_message)

    def test_feed_values_collected(self):
        """Test valid message for collected data over time."""
        reference_temperature = "T"
        value_temperature = 123
        reference_location = "LOC"
        value_location1 = "45.45,19.19"
        value_location2 = "42.42,18.18"
        reference_information = "I"
        value_information = "Hello!"

        timestamp_first = round(time.time()) * 1000
        timestamp_second = timestamp_first + 123456789

        expected_topic = self.factory.common_topic + WAPMF.FEED_VALUES
        expected_payload = json.dumps(
            [
                {
                    "timestamp": timestamp_first,
                    f"{reference_temperature}": value_temperature,
                    f"{reference_location}": value_location1,
                },
                {
                    "timestamp": timestamp_second,
                    f"{reference_location}": value_location2,
                    f"{reference_information}": value_information,
                },
            ]
        )
        expected_message = Message(expected_topic, expected_payload)

        serialized_message = self.factory.make_from_feed_values_collected(
            {
                timestamp_first: {
                    reference_temperature: value_temperature,
                    reference_location: value_location1,
                },
                timestamp_second: {
                    reference_location: value_location2,
                    reference_information: value_information,
                },
            }
        )

        self.assertEqual(expected_message, serialized_message)

    def test_feed_value_throws_on_invalid_data(self):
        """Test valid message for two string feeds."""
        self.assertRaises(
            ValueError, self.factory.make_from_feed_value, "foo", 1
        )

    def test_file_package_request(self):
        """Test message file package request."""
        file_name = "file_name"
        chunk_index = 0

        expected_topic = self.factory.common_topic + WAPMF.FILE_BINARY_REQUEST
        expected_payload = json.dumps(
            {
                "name": file_name,
                "chunkIndex": chunk_index,
            }
        )
        expected_message = Message(expected_topic, expected_payload)
        serialized_message = self.factory.make_from_package_request(
            file_name, chunk_index
        )

        self.assertEqual(expected_message, serialized_message)

    def test_file_status_ready(self):
        """Test message for file management status with file ready."""
        file_name = "file_name"
        status = FileManagementStatus(FileManagementStatusType.FILE_READY)
        expected_topic = self.factory.common_topic + WAPMF.FILE_UPLOAD_STATUS
        expected_payload = json.dumps(
            {"name": file_name, "status": status.status.value}
        )
        expected_message = Message(expected_topic, expected_payload)
        serialized_message = self.factory.make_from_file_management_status(
            status, file_name
        )

        self.assertEqual(expected_message, serialized_message)

    def test_file_status_error(self):
        """Test message for file management status with error status."""
        file_name = "file_name"
        status = FileManagementStatus(
            FileManagementStatusType.ERROR,
            FileManagementErrorType.UNKNOWN,
        )
        expected_topic = self.factory.common_topic + WAPMF.FILE_UPLOAD_STATUS
        expected_payload = json.dumps(
            {
                "name": file_name,
                "status": status.status.value,
                "error": status.error.value,
            }
        )
        expected_message = Message(expected_topic, expected_payload)
        serialized_message = self.factory.make_from_file_management_status(
            status, file_name
        )

        self.assertEqual(expected_message, serialized_message)

    def test_file_url_status_ready(self):
        """Test message for file URL management status with file ready."""
        file_name = "file_name"
        file_url = "file_url"
        status = FileManagementStatus(FileManagementStatusType.FILE_READY)
        expected_topic = (
            self.factory.common_topic + WAPMF.FILE_URL_DOWNLOAD_STATUS
        )
        expected_payload = json.dumps(
            {
                "fileUrl": file_url,
                "status": status.status.value,
                "fileName": file_name,
            }
        )
        expected_message = Message(expected_topic, expected_payload)
        serialized_message = self.factory.make_from_file_url_status(
            file_url, status, file_name
        )

        self.assertEqual(expected_message, serialized_message)

    def test_file_url_status_error(self):
        """Test message for file URL management status with error status."""
        file_url = "file_url"
        status = FileManagementStatus(
            FileManagementStatusType.ERROR,
            FileManagementErrorType.MALFORMED_URL,
        )
        expected_topic = (
            self.factory.common_topic + WAPMF.FILE_URL_DOWNLOAD_STATUS
        )
        expected_payload = json.dumps(
            {
                "fileUrl": file_url,
                "status": status.status.value,
                "error": status.error.value,
            }
        )
        expected_message = Message(expected_topic, expected_payload)
        serialized_message = self.factory.make_from_file_url_status(
            file_url, status
        )

        self.assertEqual(expected_message, serialized_message)

    def test_firmware_update_status(self):
        """Test message for firmware update status."""
        status = FirmwareUpdateStatus(FirmwareUpdateStatusType.INSTALLING)
        expected_topic = (
            self.factory.common_topic + WAPMF.FIRMWARE_UPDATE_STATUS
        )

        expected_payload = json.dumps({"status": status.status.value})
        expected_message = Message(expected_topic, expected_payload)
        serialized_message = self.factory.make_from_firmware_update_status(
            status
        )

        self.assertEqual(expected_message, serialized_message)

    def test_firmware_update_status_error(self):
        """Test message for firmware update status."""
        status = FirmwareUpdateStatus(
            FirmwareUpdateStatusType.ERROR,
            FirmwareUpdateErrorType.INSTALLATION_FAILED,
        )
        expected_topic = (
            self.factory.common_topic + WAPMF.FIRMWARE_UPDATE_STATUS
        )

        expected_payload = json.dumps(
            {"status": status.status.value, "error": status.error.value}
        )
        expected_message = Message(expected_topic, expected_payload)
        serialized_message = self.factory.make_from_firmware_update_status(
            status
        )

        self.assertEqual(expected_message, serialized_message)

    def test_time_request_message(self):
        """Test message for time request."""
        expected_topic = self.factory.common_topic + WAPMF.TIME
        expected_payload = None
        expected_message = Message(expected_topic, expected_payload)

        serialized_message = self.factory.make_time_request()

        self.assertEqual(expected_message, serialized_message)

    def test_pull_feed_values_message(self):
        """Test message for pull feed values request."""
        expected_topic = self.factory.common_topic + WAPMF.PULL_FEED_VALUES
        expected_payload = None
        expected_message = Message(expected_topic, expected_payload)

        serialized_message = self.factory.make_pull_feed_values()

        self.assertEqual(expected_message, serialized_message)

    def test_parameters_message(self):
        """Test message for push parameters value."""
        expected_topic = self.factory.common_topic + WAPMF.PARAMETERS
        values = {
            "bool_parameter": False,
            "int_parameter": 1,
            "float_parameter": 13.37,
            "string_parameter": "foo",
        }
        expected_payload = json.dumps(values)
        expected_message = Message(expected_topic, expected_payload)

        serialized_message = self.factory.make_from_parameters(values)

        self.assertEqual(expected_message, serialized_message)

    def test_pull_parameters_message(self):
        """Test message for pull parameters request."""
        expected_topic = self.factory.common_topic + WAPMF.PULL_PARAMETERS
        expected_payload = None
        expected_message = Message(expected_topic, expected_payload)

        serialized_message = self.factory.make_pull_parameters()

        self.assertEqual(expected_message, serialized_message)

    def test_feed_registration_message(self):
        """Test message for feed registration request."""
        expected_topic = self.factory.common_topic + WAPMF.FEED_REGISTRATION

        name = "test_feed_name"
        reference = "test_feed_reference"
        feed_type = FeedType.IN
        unit = Unit.CELSIUS

        expected_payload = json.dumps(
            [
                {
                    "name": name,
                    "reference": reference,
                    "type": feed_type.value,
                    "unitGuid": unit.value,
                }
            ]
        )
        expected_message = Message(expected_topic, expected_payload)

        serialized_message = self.factory.make_feed_registration(
            name, reference, feed_type, unit
        )

        self.assertEqual(expected_message, serialized_message)

    def test_feed_registration_message_custom_unit(self):
        """Test message for feed registration request with custom unit."""
        expected_topic = self.factory.common_topic + WAPMF.FEED_REGISTRATION

        name = "test_feed_name"
        reference = "test_feed_reference"
        feed_type = FeedType.IN
        unit = "USER_DEFINED_UNIT"

        expected_payload = json.dumps(
            [
                {
                    "name": name,
                    "reference": reference,
                    "type": feed_type.value,
                    "unitGuid": unit,
                }
            ]
        )
        expected_message = Message(expected_topic, expected_payload)

        serialized_message = self.factory.make_feed_registration(
            name, reference, feed_type, unit
        )

        self.assertEqual(expected_message, serialized_message)

    def test_feed_removal_message(self):
        """Test message for feed removal."""
        expected_topic = self.factory.common_topic + WAPMF.FEED_REMOVAL

        reference = "test_feed_reference"

        expected_payload = json.dumps([reference])
        expected_message = Message(expected_topic, expected_payload)

        serialized_message = self.factory.make_feed_removal(reference)

        self.assertEqual(expected_message, serialized_message)

    def test_attribute_registration_message(self):
        """Test message for attribute registration."""
        expected_topic = (
            self.factory.common_topic + WAPMF.ATTRIBUTE_REGISTRATION
        )

        name = "test_attribute_name"
        data_type = DataType.STRING
        value = "test_attribute_value"

        expected_payload = json.dumps(
            [
                {
                    "name": name,
                    "dataType": data_type.value,
                    "value": value,
                }
            ]
        )
        expected_message = Message(expected_topic, expected_payload)

        serialized_message = self.factory.make_attribute_registration(
            name, data_type, value
        )

        self.assertEqual(expected_message, serialized_message)



"""Core of this package. Wraps in all functionality."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import os
from inspect import signature
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from wolk import logger_factory
from wolk.in_memory_readings_persistence import InMemoryReadingsPersistence
from wolk.interface.connectivity_service import ConnectivityService
from wolk.interface.file_management import FileManagement
from wolk.interface.firmware_handler import FirmwareHandler
from wolk.interface.firmware_update import FirmwareUpdate
from wolk.interface.message_deserializer import MessageDeserializer
from wolk.interface.message_factory import MessageFactory
from wolk.interface.message_queue import MessageQueue
from wolk.interface.readings_persistence import ReadingsPersistence
from wolk.message_deque import MessageDeque
from wolk.model.data_delivery import DataDelivery
from wolk.model.data_type import DataType
from wolk.model.device import Device
from wolk.model.feed_type import FeedType
from wolk.model.file_management_error_type import FileManagementErrorType
from wolk.model.file_management_status import FileManagementStatus
from wolk.model.file_management_status_type import FileManagementStatusType
from wolk.model.firmware_update_error_type import FirmwareUpdateErrorType
from wolk.model.firmware_update_status import FirmwareUpdateStatus
from wolk.model.firmware_update_status_type import FirmwareUpdateStatusType
from wolk.model.message import Message
from wolk.model.unit import Unit
from wolk.mqtt_connectivity_service import MQTTConnectivityService as MQTTCS
from wolk.os_file_management import OSFileManagement
from wolk.os_firmware_update import OSFirmwareUpdate
from wolk.wolkabout_protocol_message_deserializer import (
    WolkAboutProtocolMessageDeserializer as WAPMDeserializer,
)
from wolk.wolkabout_protocol_message_factory import (
    WolkAboutProtocolMessageFactory as WAPMFactory,
)

IncomingData = List[Dict[str, Union[bool, int, float, str]]]
OutgoingDataTypes = Union[bool, int, float, str]
Reading = Tuple[str, OutgoingDataTypes]


class WolkConnect:
    """
    Exchange data with WolkAbout IoT Platform.

    :ivar connectivity_service: Means of sending/receiving data
    :vartype connectivity_service: ConnectivityService
    :ivar device: Contains device key and password
    :vartype device: Device
    :ivar file_management: File management module
    :vartype file_management: FileManagement or None
    :ivar firmware_update: Firmware update handler
    :vartype firmware_update: FirmwareUpdate or None
    :ivar logger: Logger instance issued by wolk.LoggerFactory
    :vartype logger: logging.Logger
    :ivar message_deserializer: Deserializer of inbound messages
    :vartype message_deserializer: MessageDeserializer
    :ivar message_factory: Create messages to send
    :vartype message_factory: MessageFactory
    :ivar message_queue: Store data before sending
    :vartype message_queue: MessageQueue
    :ivar readings_persistence: Store readings before sending
    :vartype readings_persistence: ReadingsPersistence
    :ivar readings_limit: Limit of readings stored in persistence
    :vartype readings_limit: int
    """

    def __init__(
        self,
        device: Device,
        host: Optional[str] = None,
        port: Optional[int] = None,
        ca_cert: Optional[str] = None,
    ):
        """
        Provide communication with WolkAbout IoT Platform.

        :param device: Contains key and password
        :type device: Device
        :param host: Host name or IP address of the remote broker
        :type host: str, optional
        :param port: Network port of the server host to connect to
        :type port: int, optional
        :param ca_cert: String path to Certificate Authority certificate file
        :type ca_cert: str, optional
        """
        logger_factory.logger_factory.set_device_key(device.key)
        self.logger = logger_factory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )
        self.logger.debug(
            f"Device: {device} ; "
            f"Host: {host} ; Port: {port} ; ca_cert: {ca_cert}"
        )

        self.device = device

        self.file_management: Optional[FileManagement] = None
        self.firmware_update: Optional[FirmwareUpdate] = None
        self.message_queue: MessageQueue = MessageDeque()
        self.message_factory: MessageFactory = WAPMFactory(device.key)
        self.message_deserializer: MessageDeserializer = WAPMDeserializer(
            self.device
        )
        self.parameters: Dict[str, Union[int, bool, float, str]] = {}
        self.readings_persistence: ReadingsPersistence = (
            InMemoryReadingsPersistence()
        )
        self.readings_limit = 500000

        wolk_ca_cert = os.path.join(os.path.dirname(__file__), "ca.crt")

        if host and port and ca_cert:
            self.connectivity_service: ConnectivityService = MQTTCS(
                device,
                self.message_deserializer.get_inbound_topics(),
                host=host,
                port=int(port),
                ca_cert=ca_cert,
            )
        elif host and port:
            self.connectivity_service = MQTTCS(
                device,
                self.message_deserializer.get_inbound_topics(),
                host=host,
                port=int(port),
            )
        else:
            # NOTE: Default to secure connection to 'demo.wolkabout.com'
            self.connectivity_service = MQTTCS(
                device,
                self.message_deserializer.get_inbound_topics(),
                port=8883,
                ca_cert=wolk_ca_cert,
            )

        self.connectivity_service.set_inbound_message_listener(
            self._on_inbound_message
        )

        self.last_platform_timestamp: Optional[int] = None

    def with_file_management(  # type: ignore
        self,
        file_directory: str,
        preferred_package_size: int = 0,
        url_downloader: Optional[Callable[[str, str], bool]] = None,
    ):
        """
        Enable file management on the device.

        :param file_directory: Directory where files are stored
        :type file_directory: str
        :param preferred_package_size: Size in kilobytes, 0 means no limit
        :type preferred_package_size: int
        :param url_downloader: Function for downloading file from URL
        :type url_downloader: Optional[Callable[[str, str], bool]]
        """
        self.logger.debug(
            f"File directory: '{file_directory}',"
            f"preferred package size: {preferred_package_size}, "
        )

        self.file_management = OSFileManagement(
            self._on_file_upload_status,
            self._on_package_request,
            self._on_file_url_status,
        )
        self.file_management.configure(
            file_directory,
            preferred_package_size,
        )

        if url_downloader is not None:
            self.file_management.set_custom_url_downloader(url_downloader)

        return self

    def with_incoming_feed_value_handler(  # type: ignore
        self, incoming_feed_value_handler: Callable[[IncomingData], None]
    ):
        """
        Enable device to respond to incoming feed value change commands.

        Commands will be delivered as a list of dictionaries that contain
        a feed_reference:value pair, and a "timestamp" field that specifies
        when this command was issued from the platform.
        The timestamp is an int representing Unix milliseconds.

        :param incoming_feed_value_handler: Handler of feed value commands
        :type incoming_feed_value_handler: Callable[[IncomingData], None]
        """
        self.logger.debug(
            f"Incoming feed value handler: {incoming_feed_value_handler}"
        )

        if not callable(incoming_feed_value_handler):
            raise ValueError(
                f"{incoming_feed_value_handler} is not a callable!"
            )
        if len(signature(incoming_feed_value_handler).parameters) != 1:
            raise ValueError(
                f"{incoming_feed_value_handler} invalid signature!"
            )
        self.incoming_feed_value_handler = incoming_feed_value_handler

        return self

    def with_firmware_update(self, firmware_handler: FirmwareHandler):  # type: ignore
        """
        Enable firmware update for device.

        Requires that file management is previously enabled on device.

        :param firmware_handler: Provide firmware version & handle installation
        :type firmware_handler: FirmwareHandler
        """
        self.logger.debug(f"Firmware handler: {firmware_handler}")
        if self.file_management is None:
            raise RuntimeError(
                "File management must be enabled before firmware update"
            )
        self.firmware_update = OSFirmwareUpdate(
            firmware_handler, self._on_firmware_update_status
        )

        return self

    def with_custom_message_queue(self, message_queue: MessageQueue):  # type: ignore
        """
        Use custom means of storing serialized messages.

        :param message_queue: Custom message queue
        :type message_queue: MessageQueue
        """
        self.logger.debug(f"Message queue: {message_queue}")
        if not isinstance(message_queue, MessageQueue):
            raise ValueError(
                "Provided message queue does not implement MessageQueue"
            )

        self.message_queue = message_queue

        return self

    def with_custom_readings_persistence(self, readings_persistence: ReadingsPersistence):  # type: ignore
        """
        Use custom means of storing readings.

        :param readings_persistence: Custom readings persistence
        :type readings_persistence: ReadingsPersistence
        """
        self.logger.debug(f"Readings persistence: {readings_persistence}")
        if not isinstance(readings_persistence, ReadingsPersistence):
            raise ValueError(
                "Provided readings persistence does not implement ReadingsPersistence"
            )

        self.readings_persistence = readings_persistence

        return self

    def set_custom_readings_persistence_limit(self, limit: int):  # type: ignore
        """
        Change the limit for readings persistence.

        :param limit: New limit for readings persistence
        :type limit: int
        """
        self.readings_limit = limit
        return self

    def with_custom_protocol(  # type: ignore
        self,
        message_factory: MessageFactory,
        message_deserializer: MessageDeserializer,
    ):
        """
        Provide a custom protocol to use for communication with the Platform.

        :param message_factory: Creator of messages to be sent to the Platform
        :type message_factory: MessageFactory
        :param message_deserializer: Deserializer of messages from the Platform
        :type message_deserializer: MessageDeserializer
        """
        self.logger.debug(
            f"Message factory: {message_factory} ; "
            f"message deserializer: {message_deserializer}"
        )
        if not isinstance(message_factory, MessageFactory):
            raise ValueError("Invalid message factory provided")
        self.message_factory = message_factory

        if not isinstance(message_deserializer, MessageDeserializer):
            raise ValueError("Invalid message deserializer provided")
        self.message_deserializer = message_deserializer

        return self

    def with_custom_connectivity(  # type: ignore
        self, connectivity_service: ConnectivityService
    ):
        """
        Provide a custom way to communicate with the Platform.

        :param connectivity_service: Custom connectivity service
        :type connectivity_service: ConnectivityService
        """
        self.logger.debug(f"Connectivity service: {connectivity_service}")
        if not isinstance(connectivity_service, ConnectivityService):
            raise ValueError("Invalid connectivity service provided")
        self.connectivity_service = connectivity_service
        self.connectivity_service.set_inbound_message_listener(
            self._on_inbound_message
        )

        return self

    def connect(self) -> None:
        """
        Connect the device to the WolkAbout IoT Platform.

        If the connection is made, then it also sends information
        about list of files present on device, current firmware version
        and the result of the firmware update process.
        """
        self.logger.debug("Connecting")

        if self.connectivity_service.is_connected():
            self.logger.info("Already connected")
            return

        try:
            self.connectivity_service.connect()
        except Exception as exception:
            self.logger.exception(
                f"Something went wrong when trying to connect: {exception}"
            )
            return

        if self.connectivity_service.is_connected():
            parameters: dict = {}
            parameters["FILE_TRANSFER_PLATFORM_ENABLED"] = False
            parameters["FIRMWARE_UPDATE_ENABLED"] = False
            parameters["FILE_TRANSFER_URL_ENABLED"] = False
            if self.file_management:
                parameters["FILE_TRANSFER_PLATFORM_ENABLED"] = True
                parameters[
                    "FILE_TRANSFER_URL_ENABLED"
                ] = self.file_management.supports_url_download()
                parameters[
                    "MAXIMUM_MESSAGE_SIZE"
                ] = self.file_management.get_preffered_package_size()

                file_list = self.file_management.get_file_list()
                message = self.message_factory.make_from_file_list(file_list)
                if not self.connectivity_service.publish(message):
                    self.message_queue.put(message)
            if self.firmware_update:
                parameters["FIRMWARE_UPDATE_ENABLED"] = True
                current_version = self.firmware_update.get_current_version()
                parameters["FIRMWARE_VERSION"] = current_version
                # TODO: "FIRMWARE_UPDATE_CHECK_TIME"
                # TODO: "FIRMWARE_UPDATE_REPOSITORY"

                self.firmware_update.report_result()

            self.logger.debug(f"Updating device parameters with: {parameters}")
            self.parameters.update(parameters)
            self.logger.info(
                f"Publishing device parameters: {self.parameters}"
            )
            message = self.message_factory.make_from_parameters(
                self.parameters
            )
            if not self.connectivity_service.publish(message):
                self.message_queue.put(message)

            if self.device.data_delivery == DataDelivery.PULL:
                self.pull_parameters()
                self.pull_feed_values()

    def disconnect(self) -> None:
        """Disconnect the device from WolkAbout IoT Platform."""
        if not self.connectivity_service.is_connected():
            return
        self.logger.debug("Disconnecting")
        self.connectivity_service.disconnect()

    def add_feed_value(
        self,
        reading: Union[Reading, List[Reading]],
        timestamp: Optional[int] = None,
    ) -> None:
        """
        Place a feed value reading into storage.

        A reading is identified by a unique feed reference string and
        the current value of the feed.

        This reading can either be passed as a tuple of (reference, value)
        for a single feed or as a list of previously mentioned tuples
        to pass multiple feed readings at once.

        A Unix epoch timestamp in milliseconds as int can be provided to
        denote when the reading occurred. By default, the current system
        provided time will be assigned to a reading.

        :param reading: Feed value reading
        :type reading: Union[Reading, List[Reading]]
        :param timestamp: Unix timestamp. Defaults to system time.
        :type timestamp: Optional[int]
        """
        self.logger.debug(
            f"Adding feed value: reading: {reading}, timestamp = {timestamp}"
        )

        # Check the limit first
        existing_readings = self.readings_persistence.obtain_readings_count()
        new_readings = 1 if isinstance(reading, tuple) else len(reading)
        if existing_readings + new_readings >= self.readings_limit:
            # Create a message from existing readings, store in message queue, and then clean persistence
            message = self.message_factory.make_from_feed_values_collected(
                self.readings_persistence.obtain_readings()
            )
            self.message_queue.put(message)
            self.readings_persistence.clear_readings()

        self.readings_persistence.store_reading(reading, timestamp)

    def add_feed_value_separated(
        self,
        reading: Union[Reading, List[Reading]],
        timestamp: Optional[int] = None,
    ) -> None:
        """
        Place a feed value reading into storage.

        A reading is identified by a unique feed reference string and
        the current value of the feed.

        This reading can either be passed as a tuple of (reference, value)
        for a single feed or as a list of previously mentioned tuples
        to pass multiple feed readings at once.

        A Unix epoch timestamp in milliseconds as int can be provided to
        denote when the reading occurred. By default, the current system
        provided time will be assigned to a reading.

        The separated variant will ensure that these reading values get sent
        as a separate message, independent of any other feed values that
        have been added to the object.

        :param reading: Feed value reading
        :type reading: Union[Reading, List[Reading]]
        :param timestamp: Unix timestamp. Defaults to system time.
        :type timestamp: Optional[int]
        """
        self.logger.debug(
            f"Adding feed value sealed: reading: {reading}, timestamp = {timestamp}"
        )

        message = self.message_factory.make_from_feed_value(reading, timestamp)
        # NOTE: if device is PUSH, do we try to publish instantly?
        self.message_queue.put(message)

    def publish(self) -> None:
        """Publish all currently stored messages to WolkAbout IoT Platform."""
        self.logger.debug("Publishing")
        if not self.connectivity_service.is_connected():
            self.logger.warning("Not connected, unable to publish messages")
            return

        saved_readings = len(self.readings_persistence.obtain_readings())
        if saved_readings > 0:
            readings_message = (
                self.message_factory.make_from_feed_values_collected(
                    self.readings_persistence.obtain_readings()
                )
            )
            if readings_message is not None:
                if self.connectivity_service.publish(readings_message):
                    self.readings_persistence.clear_readings()
                else:
                    self.logger.warning(
                        f"Failed to publish message: {readings_message}"
                    )

        while True:
            message = self.message_queue.peek()
            if message is None:
                break

            if self.connectivity_service.publish(message):
                self.message_queue.get()
            else:
                self.logger.warning(f"Failed to publish message: {message}")
                break
        self.logger.debug("Publishing ended")

    def pull_parameters(self) -> None:
        """Issue a message to pull commanded feed values."""
        self.logger.info("Sending pull parameters request")
        if self.device.data_delivery == DataDelivery.PUSH:
            self.logger.warning(
                "Always connected devices (PUSH) do not"
                " need to issue pull commands."
            )
            return
        if not self.connectivity_service.is_connected():
            self.logger.warning(
                "Not connected - not sending pull parameters message"
            )
            return

        message = self.message_factory.make_pull_parameters()

        if not self.connectivity_service.publish(message):
            self.logger.warning(f"Failed to publish message: {message}")

    def pull_feed_values(self) -> None:
        """Issue a message to pull commanded feed values."""
        self.logger.info("Sending pull feed values request")
        if self.device.data_delivery == DataDelivery.PUSH:
            self.logger.warning(
                "Always connected devices (PUSH) do not"
                " need to issue pull commands."
            )
            return
        if not self.connectivity_service.is_connected():
            self.logger.warning(
                "Not connected - not sending pull feed values message"
            )
            return

        message = self.message_factory.make_pull_feed_values()

        if not self.connectivity_service.publish(message):
            self.logger.warning(f"Failed to publish message: {message}")

    def request_timestamp(self) -> Optional[int]:
        """
        Return last received timestamp from Platform.

        If the device didn't connect at least once,
        then this will return None.

        :returns: UTC timestamp in milliseconds
        :rtype: int or None
        """
        if self.last_platform_timestamp is None:
            self.logger.warning("No timestamp available yet, will return None")
        if not self.connectivity_service.is_connected():
            self.logger.warning("Not connected, not requesting new time")
            return self.last_platform_timestamp

        message = self.message_factory.make_time_request()

        if not self.connectivity_service.publish(message):
            self.logger.warning(f"Failed to publish message: {message}")

        return self.last_platform_timestamp

    def register_feed(
        self,
        name: str,
        reference: str,
        feed_type: FeedType,
        unit: Union[Unit, str],
    ) -> None:
        """
        Register a new feed for the device.

        Feed is identified by name, unique reference, type (in or in/out)
        and unit; Where unit is either a default available unit listed in
        Unit enumeration, or a custom user defined unit that should be passed
        as a string value.

        :param name: Feed name
        :type name: str
        :param reference: Unique identifier
        :type reference: str
        :param feed_type: Is the feed one or two-way communication
        :type feed_type: FeedType
        :param unit: Unit used to measure this feed
        :type unit: Union[Unit, str]
        """
        self.logger.debug(
            "Register feed called with: "
            f"name='{name}', "
            f"reference='{reference}', "
            f"feed_type={feed_type}, "
            f"unit='{unit}'"
        )

        if not isinstance(unit, Unit):
            self.logger.warning(
                "Registering feed with user defined unit "
                f"'{unit}'. "
                "If this unit is not present on the platform, "
                "this request will fail."
            )

        message = self.message_factory.make_feed_registration(
            name, reference, feed_type, unit
        )

        if not self.connectivity_service.is_connected():
            self.logger.warning(
                "Not connected - not sending register feed request"
            )
            self.message_queue.put(message)
            return

        if not self.connectivity_service.publish(message):
            self.logger.warning(f"Failed to publish message: {message}")
            self.message_queue.put(message)

    def remove_feed(self, reference: str) -> None:
        """
        Remove a feed from the device.

        :param reference: Unique identifier
        :type reference: str
        """
        self.logger.debug(f"Remove feed called with: reference='{reference}'")
        message = self.message_factory.make_feed_removal(reference)

        if not self.connectivity_service.is_connected():
            self.logger.warning(
                "Not connected - not sending remove feed request"
            )
            self.message_queue.put(message)
            return

        if not self.connectivity_service.publish(message):
            self.logger.warning(f"Failed to publish message: {message}")
            self.message_queue.put(message)

    def register_attribute(
        self, name: str, data_type: DataType, value: str
    ) -> None:
        """
        Register an attribute for the device.

        The attribute name must be unique per device.
        All attributes created by a device are always required and read-only.
        If an attribute with the given name already exists,
        the value will be updated.

        :param name: Unique attribute name
        :type name: str
        :param data_type: Data type this attribute will hold
        :type data_type: DataType
        :param value: Value of the attribute
        :type value: str
        """
        self.logger.debug(
            "Register attribute called with: "
            f"name='{name}', "
            f"data_type={data_type}, "
            f"value='{value}'"
        )

        message = self.message_factory.make_attribute_registration(
            name, data_type, value
        )

        if not self.connectivity_service.is_connected():
            self.logger.warning(
                "Not connected - not sending register attribute request"
            )
            self.message_queue.put(message)
            return

        if not self.connectivity_service.publish(message):
            self.logger.warning(f"Failed to publish message: {message}")
            self.message_queue.put(message)

    def _on_inbound_message(self, message: Message) -> None:
        """
        Handle inbound messages.

        :param message: message received from the Platform
        :type message: Message
        """
        if "binary" in message.topic:
            self.logger.debug(
                f"Received message: {message.topic} , "
                f"{len(message.payload)}"
            )
        else:
            self.logger.debug(f"Received message: {message}")

        if self.message_deserializer.is_parameters(message):
            # TODO: parameters handle
            # NOTE: "FIRMWARE_UPDATE_CHECK_TIME"
            # NOTE: Other optional parameters?
            parameters = self.message_deserializer.parse_parameters(message)
            self.logger.info(f"TODO: parameters update: {parameters}")
            self.parameters.update(parameters)
            return

        if self.message_deserializer.is_feed_values(message):
            if not self.incoming_feed_value_handler:
                self.logger.warning(
                    "Received incoming feed values,"
                    " but no handler has been set!"
                )
                return

            feed_values = self.message_deserializer.parse_feed_values(message)
            if not feed_values:
                self.logger.warning(
                    "Failed to parse incoming feed values, ignoring"
                )
                return

            self.logger.info(
                f"Calling incoming feed value handler with: {feed_values}"
            )
            self.incoming_feed_value_handler(feed_values)
            return

        if self.message_deserializer.is_time_response(message):
            timestamp = self.message_deserializer.parse_time_response(message)
            self.logger.info(
                "Updating last platform timestamp "
                f"from {self.last_platform_timestamp}"
                f" to {timestamp}"
            )
            self.last_platform_timestamp = timestamp
            return

        if self.message_deserializer.is_file_management_message(message):
            self._on_file_management_message(message)
            return

        if self.message_deserializer.is_firmware_message(message):
            self._on_firmware_message(message)
            return

        self.logger.warning(f"Received unknown message: {message}")

    def _on_file_management_message(self, message: Message) -> None:

        if not self.file_management:
            self.logger.warning(
                f"Received unexpected file management message: {message}"
            )
            status = FileManagementStatus(
                FileManagementStatusType.ERROR,
                FileManagementErrorType.TRANSFER_PROTOCOL_DISABLED,
            )
            message = self.message_factory.make_from_file_management_status(
                status, ""
            )
            if not self.connectivity_service.publish(message):
                self.message_queue.put(message)
            return

        if self.message_deserializer.is_file_upload_initiate(message):
            name, size, fhash = self.message_deserializer.parse_file_initiate(
                message
            )
            if name != "":  # ignore invalid messages
                self.file_management.handle_upload_initiation(
                    name, size, fhash
                )
            return

        if self.message_deserializer.is_file_binary_response(message):
            package = self.message_deserializer.parse_file_binary(message)
            self.file_management.handle_file_binary_response(package)
            return

        if self.message_deserializer.is_file_upload_abort(message):
            self.file_management.handle_file_upload_abort()
            return

        if self.message_deserializer.is_file_url_abort(message):
            self.file_management.handle_file_upload_abort()
            return

        if self.message_deserializer.is_file_url_initiate(message):
            file_url = self.message_deserializer.parse_file_url(message)
            if file_url != "":  # ignore invalid messages
                self.file_management.handle_file_url_download_initiation(
                    file_url
                )
            return

        if self.message_deserializer.is_file_list(message):
            file_list = self.file_management.get_file_list()
            message = self.message_factory.make_from_file_list(file_list)
            if not self.connectivity_service.publish(message):
                self.message_queue.put(message)
            return

        if self.message_deserializer.is_file_delete_command(message):
            file_names = self.message_deserializer.parse_file_delete_command(
                message
            )
            if file_names:  # ignore invalid messages
                self.file_management.handle_file_delete(file_names)
                file_list = self.file_management.get_file_list()
                message = self.message_factory.make_from_file_list(file_list)
                if not self.connectivity_service.publish(message):
                    self.message_queue.put(message)
            return

        if self.message_deserializer.is_file_purge_command(message):
            self.file_management.handle_file_purge()
            file_list = self.file_management.get_file_list()
            message = self.message_factory.make_from_file_list(file_list)
            if not self.connectivity_service.publish(message):
                self.message_queue.put(message)
            return

        self.logger.warning(f"Received unknown message: {message}")

    def _on_firmware_message(self, message: Message) -> None:
        if not self.firmware_update:
            self.logger.warning(
                f"Received unexpected firmware update message: {message}"
            )
            firmware_status = FirmwareUpdateStatus(
                FirmwareUpdateStatusType.ERROR,
                FirmwareUpdateErrorType.UNKNOWN,
            )
            message = self.message_factory.make_from_firmware_update_status(
                firmware_status
            )
            if not self.connectivity_service.publish(message):
                self.message_queue.put(message)
            return

        if self.message_deserializer.is_firmware_install(message):
            file_name = self.message_deserializer.parse_firmware_install(
                message
            )
            file_path = self.file_management.get_file_path(file_name)
            if not file_path:
                self.logger.error(
                    f"Specified file not found on device! Message: {message}"
                )
                firmware_status = FirmwareUpdateStatus(
                    FirmwareUpdateStatusType.ERROR,
                    FirmwareUpdateErrorType.UNKNOWN_FILE,
                )
                message = (
                    self.message_factory.make_from_firmware_update_status(
                        firmware_status
                    )
                )
                if not self.connectivity_service.publish(message):
                    self.message_queue.put(message)
                return
            self.firmware_update.handle_install(file_path)
            return

        if self.message_deserializer.is_firmware_abort(message):
            self.firmware_update.handle_abort()
            return

        self.logger.warning(f"Received unknown firmware message: {message}")

    def _on_package_request(self, file_name: str, chunk_index: int) -> None:
        """
        Handle file transfer package requests.

        :param file_name: The name of the file to which the chunk belongs
        :type file_name: str
        :param chunk_index: The index of the requested chunk
        :type chunk_index: int
        """
        message = self.message_factory.make_from_package_request(
            file_name, chunk_index
        )
        if not self.connectivity_service.publish(message):
            self.message_queue.put(message)

    def _on_firmware_update_status(self, status: FirmwareUpdateStatus) -> None:
        """
        Report firmware update status to WolkAbout IoT Platform.

        :param status: The status to be reported to the Platform
        :type status: FirmwareUpdateStatus
        """
        message = self.message_factory.make_from_firmware_update_status(status)
        if self.connectivity_service.is_connected():
            if not self.connectivity_service.publish(message):
                self.message_queue.put(message)
        else:
            self.message_queue.put(message)

        if (
            status.status == FirmwareUpdateStatusType.SUCCESS
            and self.firmware_update
        ):
            version = self.firmware_update.get_current_version()
            self.parameters.update({"FIRMWARE_VERSION": version})
            if self.connectivity_service.is_connected():
                message = self.message_factory.make_from_parameters(
                    self.parameters
                )
                if not self.connectivity_service.publish(message):
                    self.message_queue.put(message)

    def _on_file_upload_status(
        self, file_name: str, status: FileManagementStatus
    ) -> None:
        """
        Report file upload status to WolkAbout IoT Platform.

        :param file_name: File being transferred
        :type file_name: str
        :param status: Description of current status
        :type status: FileManagementStatus
        """
        message = self.message_factory.make_from_file_management_status(
            status, file_name
        )
        if not self.connectivity_service.publish(message):
            self.message_queue.put(message)
        if (
            status.status == FileManagementStatusType.FILE_READY
            and self.file_management
        ):
            file_list = self.file_management.get_file_list()
            message = self.message_factory.make_from_file_list(file_list)
            if not self.connectivity_service.publish(message):
                self.message_queue.put(message)

    def _on_file_url_status(
        self,
        file_url: str,
        status: FileManagementStatus,
        file_name: Optional[str] = None,
    ) -> None:
        """
        Report file URL transfer status to WolkAbout IoT Platform.

        :param file_url: URL from where to download file
        :type file_url: str
        :param status: Description of current status
        :type status: FileManagementStatus
        :param file_name: Name of downloaded file
        :type file_name: Optional[str]
        """
        message = self.message_factory.make_from_file_url_status(
            file_url, status, file_name
        )

        if not self.connectivity_service.publish(message):
            self.message_queue.put(message)

        if file_name and self.file_management:
            file_list = self.file_management.get_file_list()
            message = self.message_factory.make_from_file_list(file_list)
            if not self.connectivity_service.publish(message):
                self.message_queue.put(message)



"""Persistence storing readings for grouped publishing."""
#   Copyright 2023 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from abc import ABC
from abc import abstractmethod
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from wolk.interface.message_factory import OutgoingDataTypes
from wolk.interface.message_factory import Reading


class ReadingsPersistence(ABC):
    """Persistence storing readings for grouped publishing."""

    @abstractmethod
    def store_reading(
        self,
        reading: Union[Reading, List[Reading]],
        timestamp: Optional[int] = None,
    ) -> bool:
        """
        Store the reading in persistence.

        :param reading: The reading(s) that should be added into persistence for specific timestamp.
        :type reading: Union[Reading, List[Reading]]
        :param timestamp: The time at which these readings were current/up-to-date data.
        :type timestamp: Optional[int]
        :returns: Whether the data was successfully stored.
        :rtype: bool
        """
        raise NotImplementedError()

    @abstractmethod
    def obtain_readings(self) -> Dict[int, Dict[str, OutgoingDataTypes]]:
        """
        Obtain all the readings from the persistence.

        :returns: The dictionary containing all readings, grouped by timestamp.
        :rtype: Dict[int, List[Reading]]
        """
        raise NotImplementedError()

    @abstractmethod
    def obtain_readings_count(self) -> int:
        """
        Obtain the count of all readings stored in persistence.

        :returns: The count of all readings.
        :rtype: int
        """
        raise NotImplementedError()

    @abstractmethod
    def clear_readings(self) -> bool:
        """
        Clear the entire set of readings currently stored in the persistence.

        :returns: Whether the persistence was successfully cleared.
        :rtype: bool
        """
        raise NotImplementedError()



"""OS File Management module."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import hashlib
import math
import os
import shutil
from inspect import signature
from tempfile import NamedTemporaryFile
from threading import Timer
from typing import Callable
from typing import Dict
from typing import IO
from typing import List
from typing import Optional
from typing import Union
from urllib.parse import urlparse

import requests

from wolk import logger_factory
from wolk.interface.file_management import FileManagement
from wolk.model.file_management_error_type import FileManagementErrorType
from wolk.model.file_management_status import FileManagementStatus
from wolk.model.file_management_status_type import FileManagementStatusType
from wolk.model.file_transfer_package import FileTransferPackage


class OSFileManagement(FileManagement):
    """
    File transfer manager.

    Enables device to transfer files from WolkAbout IoT Platform
    package by package or/and URL download as well as report list of files
    currently on device and delete them on request.
    """

    def __init__(
        self,
        status_callback: Callable[[str, FileManagementStatus], None],
        packet_request_callback: Callable[[str, int], None],
        url_status_callback: Callable[
            [str, FileManagementStatus, Optional[str]], None
        ],
    ) -> None:
        """
        Enable file management for device.

        :param status_callback: Reporting current file transfer status
        :type status_callback: Callable[[FileManagementStatus], None]
        :param packet_request_callback: Request file packet from Platform
        :type packet_request_callback: Callable[[str, int], None]
        :param url_status_callback: Report file URL download status
        :type url_status_callback: Callable[[str, FileManagementStatus, Optional[str]], None]
        """
        self.logger = logger_factory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )

        self.status_callback = status_callback
        self.packet_request_callback = packet_request_callback
        self.url_status_callback = url_status_callback
        self.url_downloader = self.url_download
        self.preferred_package_size = 0
        self.file_directory = ""
        self.current_status: Optional[FileManagementStatus] = None
        self.max_retries = 3
        self.next_package_index: Optional[int] = None
        self.expected_number_of_packages: Optional[int] = None
        self.retry_count = 0
        self.request_timeout: Optional[Timer] = None
        self.last_package_hash = 32 * b"\x00"
        self.temp_file: Optional[IO[bytes]] = None
        self.file_name: Optional[str] = None
        self.file_size: Optional[int] = None
        self.file_hash: Optional[str] = None
        self.file_url: Optional[str] = None

    def configure(
        self,
        file_directory: str,
        preferred_package_size: int = 0,
    ) -> None:
        """
        Configure options for file management module.

        :param file_directory: Path to where files are stored
        :type file_directory: str
        :param preferred_package_size: Size in kilobytes, 0 means no limit
        :type preferred_package_size: int
        """
        self.file_directory = file_directory
        self.preferred_package_size = preferred_package_size

        if not os.path.exists(os.path.abspath(self.file_directory)):
            os.makedirs(os.path.abspath(self.file_directory))

    def get_preffered_package_size(self) -> int:
        """
        Return preffered package size for file transfer.

        :returns: preferred_package_size
        :rtype: int
        """
        return self.preferred_package_size

    def set_custom_url_downloader(
        self, downloader: Callable[[str, str], bool]
    ) -> None:
        """
        Set the URL file downloader to a custom implementation.

        Default implementation uses `requests` and is available as a
        static method within this class.

        :param downloader: Function that will download the file from the URL
        :type downloader: Callable[[[Arg(str, 'file_url'), Arg(str, 'file_path')], bool]
        """
        self.logger.debug(f"Setting custom URL downloader to {downloader}")
        if not callable(downloader):
            self.logger.warning(f"{downloader} is not a callable!")  # type: ignore
            return
        if len(signature(downloader).parameters) != 2:
            self.logger.warning(f"{downloader} invalid signature!")
            return
        self.url_downloader = downloader  # type: ignore

    def supports_url_download(self) -> bool:
        """
        Return if the file management module supports URL download.

        :returns: supports_url_download
        :rtype: bool
        """
        return self.url_downloader is not None

    def handle_upload_initiation(
        self, file_name: str, file_size: int, file_hash: str
    ) -> None:
        """
        Start making package requests and set status to file transfer.

        :param file_name: File name
        :type file_name: str
        :param file_size: Size in bytes
        :type file_size: int
        :param file_hash: MD5 hash of file
        :type file_hash: str
        """
        if self.current_status is not None:
            self.logger.warning(
                "Not in idle state, ignoring file upload initiation"
            )
            return

        self.logger.info("Starting file transfer")
        self.logger.info(
            f"File name: {file_name} ; "
            f"File size: {file_size} ; "
            f"File hash: {file_hash}"
        )

        self.expected_number_of_packages = (
            math.ceil(file_size / (self.preferred_package_size * 1024))
            if self.preferred_package_size != 0
            else 1
        )

        if os.path.exists(
            os.path.join(os.path.abspath(self.file_directory), file_name)
        ):
            valid_file = False

            try:
                with open(
                    os.path.join(
                        os.path.abspath(self.file_directory), file_name
                    ),
                    "rb",
                ) as existing_file:
                    data = existing_file.read()
                    if not data:
                        self.logger.error("Read data returned None!")
                        return
                    calculated_hash = hashlib.md5(data).hexdigest()
            except Exception as e:
                self.logger.exception(f"Error opening existing file: {e}")
                self.current_status = FileManagementStatus(
                    FileManagementStatusType.ERROR,
                    FileManagementErrorType.FILE_SYSTEM_ERROR,
                )
                self.status_callback(file_name, self.current_status)
                self.handle_file_upload_abort()
                return

            valid_file = file_hash == calculated_hash

            if valid_file:
                self.logger.info(
                    "File requested for transfer already on device"
                )
                self.current_status = FileManagementStatus(
                    FileManagementStatusType.FILE_READY
                )
                self.status_callback(file_name, self.current_status)

                self.current_status = None
                self.last_package_hash = 32 * b"\x00"
                return

            else:
                self.logger.warning(
                    "File requested for transfer already on device, "
                    "but the hashes do not match!"
                )
                self.current_status = FileManagementStatus(
                    FileManagementStatusType.ERROR,
                    FileManagementErrorType.FILE_HASH_MISMATCH,
                )
                self.status_callback(file_name, self.current_status)

                self.current_status = None
                self.last_package_hash = 32 * b"\x00"
                return

        self.current_status = FileManagementStatus(
            FileManagementStatusType.FILE_TRANSFER
        )
        self.temp_file = NamedTemporaryFile(mode="r+b", delete=False)
        self.file_name = file_name
        self.file_size = file_size
        self.file_hash = file_hash
        self.next_package_index = 0
        self.retry_count = 0

        self.logger.info(f"Initializing file transfer: '{file_name}'")
        self.logger.info(
            "File is expected to be received in "
            + f"{self.expected_number_of_packages}"
            + " packages"
        )
        self.status_callback(self.file_name, self.current_status)

        self.packet_request_callback(self.file_name, self.next_package_index)

        self.request_timeout = Timer(60.0, self._timeout)
        self.request_timeout.start()

    def handle_file_upload_abort(self) -> None:
        """Abort file upload and revert to idle status."""
        self.logger.info("Aborting file upload.")
        if self.temp_file:
            self.temp_file.close()

        self.temp_file = None
        self.file_name = None
        self.file_size = None
        self.file_hash = None
        self.current_status = None
        self.next_package_index = None
        self.expected_number_of_packages = None
        self.retry_count = 0
        self.request_timeout = None
        self.last_package_hash = 32 * b"\x00"

    def handle_file_binary_response(
        self, package: FileTransferPackage
    ) -> None:
        """
        Validate received package and store or use callback to request again.

        :param package: Package of file being transfered.
        :type package: FileTransferPackage
        """
        self.logger.debug(
            f"Previous hash: {package.previous_hash!r} ; "
            f"Data size: {len(package.data)} ; "
            f"Current hash: {package.current_hash!r}"
        )

        if self.current_status is None:
            self.logger.warning(
                "File transfer not in progress, ignoring package"
            )
            return

        if self.request_timeout:
            self.request_timeout.cancel()

        valid_package = False

        if (
            len(package.previous_hash)
            + len(package.data)
            + len(package.current_hash)
            < 65
        ):
            self.logger.error("Received package size too small!")
        else:
            data_hash = hashlib.sha256(package.data).digest()
            if package.current_hash != data_hash:
                self.logger.error(
                    f"Data hash '{data_hash!r}' does not match "
                    f"expected hash '{package.current_hash!r}' !"
                )
            else:
                self.logger.debug(
                    f"Valid package #{self.next_package_index} received"
                )
                valid_package = True

        if (
            not valid_package
            or self.last_package_hash != package.previous_hash
        ):
            self.logger.warning("Received invalid file package")
            self.retry_count += 1

            if self.retry_count >= self.max_retries:
                self.logger.error(
                    "Retry count exceeded, aborting file transfer"
                )
                self.current_status = FileManagementStatus(
                    FileManagementStatusType.ERROR,
                    FileManagementErrorType.RETRY_COUNT_EXCEEDED,
                )
                self.status_callback(self.file_name, self.current_status)
                self.handle_file_upload_abort()
                return

            self.logger.info(
                f"Requesting package #{self.next_package_index} again"
            )
            self.packet_request_callback(
                self.file_name, self.next_package_index
            )

            self.request_timeout = Timer(60.0, self._timeout)
            self.request_timeout.start()
            return

        self.last_package_hash = package.current_hash

        try:
            self.temp_file.write(package.data)
            self.temp_file.flush()
            os.fsync(self.temp_file)
        except Exception:
            self.logger.error(
                "Failed to write package, aborting file transfer"
            )
            self.current_status = FileManagementStatus(
                FileManagementStatusType.ERROR,
                FileManagementErrorType.FILE_SYSTEM_ERROR,
            )
            self.status_callback(self.file_name, self.current_status)
            self.handle_file_upload_abort()
            return

        self.next_package_index += 1

        if self.next_package_index < self.expected_number_of_packages:
            self.logger.debug(
                f"Stored package, requesting "
                f"#{self.next_package_index}/"
                f"#{self.expected_number_of_packages}"
            )
            self.packet_request_callback(
                self.file_name,
                self.next_package_index,
            )

            self.current_status = FileManagementStatus(
                FileManagementStatusType.FILE_TRANSFER
            )

            self.request_timeout = Timer(60.0, self._timeout)
            self.request_timeout.start()
            return

        valid_file = False

        try:
            self.temp_file.seek(0)
            data = self.temp_file.read()
            if not data:
                self.logger.error("Read data from temp file returned None!")
            calculated_hash = hashlib.md5(data).hexdigest()
        except Exception as e:
            self.logger.exception(f"Error while working with temp file: {e}")
            self.current_status = FileManagementStatus(
                FileManagementStatusType.ERROR,
                FileManagementErrorType.FILE_SYSTEM_ERROR,
            )
            self.status_callback(self.file_name, self.current_status)
            self.handle_file_upload_abort()
            return

        valid_file = self.file_hash == calculated_hash

        # FIXME: remove after bugfix:
        if not valid_file:
            self.logger.warning(
                "Hash mismatch!\n"
                + f"Received: {self.file_hash}\n"
                + f"Calculated: {calculated_hash}"
            )
            if len(self.file_hash) > 32:
                self.logger.warning("Known bug, ignoring hash diff")
                valid_file = True

        if not valid_file:
            self.logger.error("Invalid file - File hash does not match")
            self.current_status.status = FileManagementStatusType.ERROR
            self.current_status.error = (
                FileManagementErrorType.FILE_HASH_MISMATCH
            )
            self.status_callback(self.file_name, self.current_status)
            self.handle_file_upload_abort()
            return

        if not os.path.exists(os.path.abspath(self.file_directory)):
            os.makedirs(os.path.abspath(self.file_directory))

        file_path = os.path.join(
            os.path.abspath(self.file_directory),
            self.file_name,
        )

        try:
            self.logger.info(
                f"Finalizing file transfer for '{self.file_name}'"
            )
            shutil.copy2(os.path.realpath(self.temp_file.name), file_path)
            self.temp_file.close()
        except Exception as e:
            self.logger.exception(f"Error when storing file: {e}")

        if not os.path.exists(file_path):
            self.logger.error(f"File failed to store to at: {file_path}")
            self.current_status = FileManagementStatus(
                FileManagementStatusType.ERROR,
                FileManagementErrorType.FILE_SYSTEM_ERROR,
            )
            self.status_callback(self.file_name, self.current_status)
            self.handle_file_upload_abort()
            return

        self.logger.info(f"Received file '{self.file_name}'")
        self.current_status = FileManagementStatus(
            FileManagementStatusType.FILE_READY
        )
        self.status_callback(self.file_name, self.current_status)

        self.current_status = None
        self.retry_count = 0
        self.last_package_hash = 32 * b"\x00"

    def handle_file_url_download_initiation(self, file_url: str) -> None:
        """
        Start file transfer from specified URL.

        :param file_url: URL from where to download file
        :type file_url: str
        """
        if self.current_status is not None:
            self.logger.warning(
                "Not in idle state, ignoring file upload initiation"
            )
            return

        if self.url_downloader is None:
            self.logger.warning(
                "Received URL download, but no downloader is set!"
            )
            self.current_status = FileManagementStatus(
                FileManagementStatusType.ERROR,
                FileManagementErrorType.TRANSFER_PROTOCOL_DISABLED,
            )
            self.url_status_callback(file_url, self.current_status, None)
            self.handle_file_upload_abort()
            return

        if not bool(urlparse(file_url).scheme):
            self.logger.error(f"Received URL '{file_url}' is not valid!")
            self.current_status = FileManagementStatus(
                FileManagementStatusType.ERROR,
                FileManagementErrorType.MALFORMED_URL,
            )
            self.url_status_callback(file_url, self.current_status, None)
            self.handle_file_upload_abort()
            return

        self.file_url = file_url
        self.file_name = self.file_url.split("/")[-1]
        file_path = os.path.join(
            os.path.abspath(self.file_directory), self.file_name
        )
        self.current_status = FileManagementStatus(
            FileManagementStatusType.FILE_TRANSFER
        )
        self.url_status_callback(file_url, self.current_status, self.file_name)

        self.url_downloader(file_url, file_path)

        if not os.path.exists(file_path):
            self.logger.error(f"File failed to store to at: {file_path}")
            self.current_status = FileManagementStatus(
                FileManagementStatusType.ERROR,
                FileManagementErrorType.FILE_SYSTEM_ERROR,
            )
            self.url_status_callback(
                file_url, self.current_status, self.file_name
            )
            self.handle_file_url_download_abort()
            return

        self.logger.info(f"File obtained from URL: '{file_url}'")
        self.current_status = FileManagementStatus(
            FileManagementStatusType.FILE_READY
        )
        self.url_status_callback(file_url, self.current_status, self.file_name)

        self.current_status = None

    def handle_file_url_download_abort(self) -> None:
        """Abort file URL download."""
        self.logger.info("Aborting URL download")
        self.file_url = None
        self.file_name = None
        self.current_status = None

    def get_file_list(self) -> List[Dict[str, Union[str, int]]]:
        """
        Return list of files present on device.

        Each list item is a dictionary that contains the name of the file,
        its size in bytes, and a MD5 checksum of the file.

        :returns: file_list
        :rtype: List[Dict[str, Union[str, int]]]
        """
        files = os.listdir(os.path.abspath(self.file_directory))
        files_list = []

        for item in files:
            file_path = os.path.join(
                os.path.abspath(self.file_directory), item
            )
            if not os.path.isfile(file_path) or item.startswith("."):
                continue
            with open(file_path, "rb") as file:
                data = file.read()
                hash = hashlib.md5(data).hexdigest()
            size = os.path.getsize(file_path)
            files_list.append({"name": item, "size": size, "hash": hash})

        self.logger.debug(f"Files on device: {files_list}")
        return files_list  # type: ignore

    def get_file_path(self, file_name: str) -> Optional[str]:
        """
        Return path to file if it exists.

        :param file_name: File for which to get path
        :type file_name: str
        :returns: file_path
        :rtype: Optional[str]
        """
        self.logger.debug(f"Get file path for file: {file_name}")
        file_path = None
        for file in self.get_file_list():
            if file["name"] == file_name:
                file_path = os.path.join(
                    os.path.abspath(self.file_directory), file_name
                )
                break
        self.logger.debug(f"File path: {file_path}")
        return file_path

    def handle_file_delete(self, file_names: List[str]) -> None:
        """
        Delete files from device.

        :param file_names: Files to be deleted
        :type file_names: List[str]
        """
        self.logger.info(f"Attempting to delete files: {file_names}")
        for file in file_names:
            try:
                file_path = os.path.join(
                    os.path.abspath(self.file_directory), file
                )

                if os.path.exists(file_path):
                    os.remove(file_path)
            except Exception as e:
                self.logger.exception(f"'{file}' failed to delete: {e}")

    def handle_file_purge(self) -> None:
        """Delete all files from device."""
        for file in os.listdir(os.path.abspath(self.file_directory)):
            if not os.path.isfile(
                os.path.join(os.path.abspath(self.file_directory), file)
            ) or file.startswith("."):
                continue
            os.remove(os.path.join(os.path.abspath(self.file_directory), file))

    def _timeout(self) -> None:
        """Revert to idle state when timeout occurs."""
        self.logger.error("Timed out waiting for next package, aborting")
        self.current_status = FileManagementStatus(
            FileManagementStatusType.ERROR,
            FileManagementErrorType.UNKNOWN,
        )
        self.status_callback(self.file_name, self.current_status)
        self.handle_file_upload_abort()

    @staticmethod
    def url_download(file_url: str, file_path: str) -> bool:
        """
        Attempt to download file from specified URL.

        :param file_url: URL from which to download file
        :type file_url: str
        :param file_path: Path where to store file
        :type: file_path: str
        :returns: Successful download
        :rtype: bool
        """
        response = requests.get(file_url)
        with open(file_path, "wb") as file:
            file.write(response.content)
            file.flush()
            os.fsync(file)

        return os.path.exists(file_path)
